"""Implements the LabelledDataGroup, which allows to handle groups and
containers that can be associated with further coordinates.

This imitates the xarray selection interface and provides a uniform interface
to select data from these groups. Most importantly, it allows to combine all
the data of one group, allowing to conveniently work with heterogeneously
stored data.
"""

import logging
import warnings
from typing import Dict, List, Tuple, Union

import numpy as np
import xarray as xr

from ..abc import AbstractDataContainer
from ..containers import XrDataContainer
from ..exceptions import *
from ..tools import apply_along_axis
from ..utils import extract_coords
from ..utils.coords import TCoord, TCoordsDict, TDims
from . import OrderedDataGroup

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class LabelledDataGroup(OrderedDataGroup):
    """A group that assumes that the members it contains can be labelled
    with dimension names and coordinates.

    Such a group has the great benefit to provide a selection interface that
    works fully on the dimension labels and coordinates and can cooperate with
    the xarray selection interface, i.e. the ``sel`` and ``isel`` methods.
    """

    # Let new containers be xarray-based
    _NEW_CONTAINER_CLS = XrDataContainer

    # Configuration options for this group ....................................
    # Whether to use deep selection by default
    LDG_ALLOW_DEEP_SELECTION = True

    # The dimensions of this group, i.e. the dimensions of the space covered by
    # the members of this group.
    LDG_DIMS = tuple()

    # How to extract coordinates of the members; for available modes, see the
    # `dantro.utils.coords.extract_coords` function.
    LDG_EXTRACT_COORDS_FROM = "data"

    # Configuration for mode 'attrs' . . . . . . . . . . . . . . . . . . . . .
    LDG_COORDS_ATTR_PREFIX = "ext_coords__"
    LDG_COORDS_MODE_ATTR_PREFIX = "ext_coords_mode__"
    LDG_COORDS_MODE_DEFAULT = "scalar"
    LDG_STRICT_ATTR_CHECKING = False

    # Configuration for mode 'name' . . . . . . . . . . . . . . . . . . . . . .
    LDG_COORDS_SEPARATOR_IN_NAME = ";"

    # .........................................................................

    def __init__(
        self,
        *args,
        dims: TDims = None,
        mode: str = None,
        allow_deep_selection: bool = None,
        **kwargs,
    ):
        """Initialize a LabelledDataGroup

        Args:
            *args: Passed on to
                :py:class:`~dantro.groups.ordered.OrderedDataGroup`
            dims (TDims, optional): The dimensions associated with this group.
                If not given, will use those defined in the ``LDG_DIMS`` class
                variable. These can *not* be changed afterwards!
            mode (str, optional): By which coordiante extraction mode to get
                the coordinates from the group members. Can be ``attrs``,
                ``name``, ``data`` or anything else specified in
                :py:func:`~dantro.utils.coords.extract_coords`.
            allow_deep_selection (bool, optional): Whether to allow deep
                selection. If not given, will use the
                ``LDG_ALLOW_DEEP_SELECTION`` class variable's value. Behaviour
                can be changed via the property of the same name.
            **kwargs: Passed on to
                :py:class:`~dantro.groups.ordered.OrderedDataGroup`
        """
        # Initialize the member map, which is needed if containers are added
        # during initialization (thus invoking _add_container_callback)
        self.__member_map = None

        super().__init__(*args, **kwargs)

        self._dims = dims if dims is not None else tuple(self.LDG_DIMS)

        self._mode = self.LDG_EXTRACT_COORDS_FROM
        if mode is not None:
            self._mode = mode

        self._allow_deep_selection = self.LDG_ALLOW_DEEP_SELECTION
        if allow_deep_selection is not None:
            self._allow_deep_selection = allow_deep_selection

    # Dimension and coordinates ...............................................

    @property
    def dims(self) -> Tuple[str]:
        """The names of the group-level dimensions this group manages.

        It _may_ contain dimensions that overlap with dimension names from the
        members; this is intentional.
        """
        return self.member_map.dims

    @property
    def ndim(self) -> int:
        """The rank of the space covered by the group-level dimensions."""
        return self.member_map.ndim

    @property
    def coords(self) -> Dict[str, List[TCoord]]:
        """Returns a dict-like container of group-level coordinate values keyed
        by dimension.
        """
        return self.member_map.coords

    @property
    def shape(self) -> Tuple[int]:
        """Return the shape of the space covered by the group-level dimensions."""
        return self.member_map.shape

    # Additional properties ...................................................

    @property
    def allow_deep_selection(self) -> bool:
        """Whether deep selection is allowed."""
        return self._allow_deep_selection

    @allow_deep_selection.setter
    def allow_deep_selection(self, val: bool):
        """Change whether deep selection is allowed."""
        self._allow_deep_selection = val

    @property
    def member_map(self) -> xr.DataArray:
        """Returns an array that represents the space that the members of this
        group span, where each value (i.e. a specific coordinate combination)
        is the name of the corresponding member of this group.

        Upon first call, this is computed here. If members are added, it is
        tried to accomodate them in there; if not possible, the cache will be
        invalidated.

        The member map _may_ include empty strings, i.e. coordinate
        combinations that are not covered by any member. Also, they can contain
        duplicate names, as one member can cover multiple coordinates.

        .. note::

            The member map is invalidated when new members are added that can
            not be accomodated in it. It will be recalculated when needed.
        """
        if self.member_map_available:
            return self.__member_map

        # Member map is not available and has to be created anew.
        # First, extract the coordinates by iterating over all members and
        # aggregating their individual coordinates; sets guarantee uniqueness
        coords = {dim_name: set() for dim_name in self._dims}

        for cont_name, cont in self.items():
            cont_coords = self._get_coords_of(cont)

            for dim_name, coord_vals in cont_coords.items():
                coords[dim_name].update(coord_vals)

        # Convert to dict of lists
        coords = {dim_name: sorted(list(s)) for dim_name, s in coords.items()}

        # Now, derive the shape from the coordinates
        shape = tuple([len(coords[dim_name]) for dim_name in self._dims])

        # Create a DataArray containing empty strings, using the existing
        # dimension names (as set during initialization) and coordinates to
        # label it.
        mm_data = np.zeros(shape, dtype="object")
        mm_data.fill("")

        mm = xr.DataArray(
            data=mm_data,
            dims=self._dims,
            coords=coords,
        )

        # Iterate over members and populate the array with member names
        for cont_name, cont in self.items():
            cont_coords = self._get_coords_of(cont)
            # These coordinates describe a hypercube in coordinate space that
            # is to be associated with this container. Thus, the member map
            # should contain the name of the member for all these coordinates:
            mm.loc[cont_coords] = cont_name

        # Cache the map and return it
        self.__member_map = mm
        return mm

    @property
    def member_map_available(self) -> bool:
        """Whether the member map is available yet."""
        return self.__member_map is not None

    # Selection interface .....................................................

    def isel(
        self,
        indexers: dict = None,
        *,
        drop: bool = False,
        combination_method: str = "try_concat",
        deep: bool = None,
        **indexers_kwargs,
    ) -> xr.DataArray:
        """Return a new labelled `xr.DataArray` with an index-selected subset
        of members of this group.

        If deep selection is activated, those indexers that are not available
        in the group-managed dimensions are looked up in the members of this
        group.

        .. note::

            For data combination (via *any* ``combination_method``)
            dimensions that differ in size across group members have to be
            labelled, such that arrays can be
            `aligned <http://xarray.pydata.org/en/stable/generated/xarray.align.html>`_
            using the respective coordinates.
            See `the xarray documentation <http://xarray.pydata.org/en/stable/data-structures.html#coordinates>`_
            for more information about coordinates.

        Args:
            indexers (dict, optional): A dict with keys matching dimensions and
                values given by scalars, slices or arrays of tick indices.
                As `xr.DataArray.isel`, uses pandas-like indexing, i.e.: slices
                do not include the terminal value.
            drop (bool, optional): Drop coordinate variables instead of making
                them scalar.
            combination_method (str, optional): How to combine group-level data
                with member-level data. Ignored if data from a single group
                member is selected, i.e. no data has to be combined. Can be:

                    * ``concat``: Concatenate. This can preserve the dtype, but
                      requires that no data is missing.
                    * ``merge``: Merge, using `xarray.merge`. This leads to a
                      type conversion to ``float64``, but allows members being
                      missing or coordinates not fully filling the available
                      space.
                    * ``try_concat``: Try concatenation, fall back to merging
                      if that was unsuccessful.

            deep (bool, optional): Whether to allow deep indexing, i.e.: that
                ``indexers`` may contain dimensions that don't refer to group-
                level dimensions but to dimensions that are only availble among
                the member data. If ``None``, will use the value returned by
                the ``allow_deep_selection`` property.
            **indexers_kwargs: Additional indexers

        Returns:
            xr.DataArray: The selected data, potentially a combination of data
                on group level and member-level data.
        """
        idxrs, deep_idxrs = self._parse_indexers(
            indexers, allow_deep=deep, **indexers_kwargs
        )

        # Use the (shallow) indexers to select (by index) those members that
        # are to be combined ...
        tbc = self.member_map.isel(idxrs, drop=drop)

        # Now, combine them, potentially also applying deep indexing
        return self._combine(
            tbc,
            combination_method=combination_method,
            shallow_indexers=idxrs,
            deep_indexers=deep_idxrs,
            by_index=True,
            drop=drop,
        )

    def sel(
        self,
        indexers: dict = None,
        *,
        method: str = None,
        tolerance: float = None,
        drop: bool = False,
        combination_method: str = "try_concat",
        deep: bool = None,
        **indexers_kwargs,
    ) -> xr.DataArray:
        """Return a new labelled `xr.DataArray` with a coordinate-selected
        subset of members of this group.

        If deep selection is activated, those indexers that are not available
        in the group-managed dimensions are looked up in the members of this
        group.

        .. note::

            For data combination (via *any* ``combination_method``)
            dimensions that differ in size across group members have to be
            labelled, such that arrays can be
            `aligned <http://xarray.pydata.org/en/stable/generated/xarray.align.html>`_
            using the respective coordinates.
            See `the xarray documentation <http://xarray.pydata.org/en/stable/data-structures.html#coordinates>`_
            for more information about coordinates.


        Args:
            indexers (dict, optional): A dict with keys matching dimensions and
                values given by scalars, slices or arrays of tick labels.
                As `xr.DataArray.sel`, uses pandas-like indexing, i.e.: slices
                include the terminal value.
            method (str, optional): Method to use for inexact matches
            tolerance (float, optional): Maximum (absolute) distance between
                original and given label for inexact matches.
            drop (bool, optional): Drop coordinate variables instead of making
                them scalar.
            combination_method (str, optional): How to combine group-level data
                with member-level data. Ignored if data from a single group
                member is selected, i.e. no data has to be combined. Can be:

                    * ``concat``: Concatenate. This can preserve the dtype, but
                      requires that no data is missing.
                    * ``merge``: Merge, using `xarray.merge`. This leads to a
                      type conversion to ``float64``, but allows members being
                      missing or coordinates not fully filling the available
                      space.
                    * ``try_concat``: Try concatenation, fall back to merging
                      if that was unsuccessful.

            deep (bool, optional): Whether to allow deep indexing, i.e.: that
                ``indexers`` may contain dimensions that don't refer to group-
                level dimensions but to dimensions that are only availble among
                the member data. If ``None``, will use the value returned by
                the ``allow_deep_selection`` property.
            **indexers_kwargs: Additional indexers

        Returns:
            xr.DataArray: The selected data, potentially a combination of data
                on group level and member-level data.
        """
        idxrs, deep_idxrs = self._parse_indexers(
            indexers, allow_deep=deep, **indexers_kwargs
        )

        # Use the (shallow) indexers to select (by label) those members that
        # are to be combined ...
        tbc = self.member_map.sel(
            idxrs, method=method, tolerance=tolerance, drop=drop
        )

        # Now, combine them, potentially also applying deep indexing
        return self._combine(
            tbc,
            combination_method=combination_method,
            shallow_indexers=idxrs,
            deep_indexers=deep_idxrs,
            by_index=False,
            method=method,
            tolerance=tolerance,
            drop=drop,
        )

    # Helpers .................................................................
    # General . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    def _get_coords_of(self, obj: AbstractDataContainer) -> TCoordsDict:
        """Extract the coordinates for the given object using the
        :py:func:`~dantro.utils.coords.extract_coords` function.

        Args:
            obj (AbstractDataContainer): The object to get the coordinates of.

        Returns:
            TCoordsDict: The extracted coordinates
        """
        # Depending on the mode, compile the dict of additional parameters
        kwargs = dict()

        if self._mode == "attrs":
            kwargs["coords_attr_prefix"] = self.LDG_COORDS_ATTR_PREFIX
            kwargs["mode_attr_prefix"] = self.LDG_COORDS_MODE_ATTR_PREFIX
            kwargs["default_mode"] = self.LDG_COORDS_MODE_DEFAULT
            kwargs["strict"] = self.LDG_STRICT_ATTR_CHECKING

        elif self._mode == "name":
            kwargs["separator"] = self.LDG_COORDS_SEPARATOR_IN_NAME

        return extract_coords(obj, dims=self._dims, mode=self._mode, **kwargs)

    def _add_container_callback(self, cont: AbstractDataContainer) -> None:
        """Called by the base class after adding a container, this method
        checks whether the member map needs to be invalidated or whether the
        new container can be accomodated in it.

        If it can be accomodated, the member map will be adjusted such that for
        all coordinates associated with the given ``cont``, the member map
        points to the newly added container.

        Args:
            cont (AbstractDataContainer): The newly added container
        """
        # First, let the parent class do its thinkg
        super()._add_container_callback(cont)

        # Don't have to do anything if there is no member map yet
        if not self.member_map_available:
            return

        # There is a map. Check if it can accomodate the new container
        coords = self._get_coords_of(cont)

        try:
            self.__member_map.loc[coords] = cont.name

        except Exception:
            # Cannot accomodate it -> invalidate it
            self.__member_map = None

    # For selection . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    def _parse_indexers(
        self, indexers: dict, *, allow_deep: bool, **indexers_kwargs
    ) -> Tuple[dict, dict]:
        """Parses the given indexer arguments and split them into indexers for
        the selection of group members and deep selection.

        Args:
            indexers (dict): The indexers dict, may be empty
            allow_deep (bool): Whether to allow deep selection
            **indexers_kwargs: Additional indexers

        Returns:
            Tuple[dict, dict]: (shallow indexers, deep indexers)

        Raises:
            ValueError: If deep indexers were given but deep selection was not
                enabled
        """
        allow_deep = (
            allow_deep if allow_deep is not None else self.allow_deep_selection
        )
        idxrs = dict(**(indexers if indexers else {}), **indexers_kwargs)

        # Split by those for deep selection and those for this group
        deep_idxrs = {k: v for k, v in idxrs.items() if k not in self._dims}
        idxrs = {k: v for k, v in idxrs.items() if k in self._dims}

        if deep_idxrs and not allow_deep:
            _dim_names = ", ".join(self._dims)
            raise ValueError(
                f"Deep indexing is not allowed for {self.logstr}, but got "
                "indexers that don't match any of its dimension "
                f"names: {_dim_names}.  You can change this behavior using "
                "the allow_deep_selection property or the class "
                "variable LDG_ALLOW_DEEP_SELECTION."
            )

        return idxrs, deep_idxrs

    def _combine(
        self,
        cont_names: xr.DataArray,
        *,
        combination_method: str,
        shallow_indexers: dict,
        deep_indexers: dict,
        by_index: bool,
        **sel_kwargs,
    ) -> xr.Dataset:
        """Combine the given objects by the specified method. If deep indexers
        are given, apply the deep indexing on each of the members.

        This method receives a labelled array of container names, on which the
        selection already took place. The aim is now to align the objects these
        names refer to, including their coordinates, and thereby construct an
        array that contains both the dimensions given by the ``cont_names``
        array and each members' data dimensions.

        Available combination methods are based either on `xarray.merge`
        operations or `xarray.concat` along each dimension. For both these
        combination methods, the members of this group need to be prepared such
        that the operation can be applied, i.e.: they need to already be in an
        array capable of that operation and they need to directly or indirectly
        preserve coordinate information.

        For that purpose, an object-array is constructed that has the same
        shape as the given ``cont_names``. As the `xarray.Dataset` and
        `xarray.DataArray` types have issues with handling array-like objects
        in object arrays, this is done via a `numpy.ndarray`.

        Args:
            cont_names (xr.DataArray): The pre-selected member map object, i.e.
                a labelled array containing names of the desired members that
                are to be combined.
            combination_method (str): How to combine them: concat, try_concat,
                or merge. Concatenation will allow preserving the dtype of the
                underlying data.
            shallow_indexers (dict): Indexer arguments that were used for the
                group member selection.
            deep_indexers (dict): Indexer arguments for deep selection to be
                done before combination.
            by_index (bool): Whether the deep indexing should take place by
                index; if False, will use label-based selection.
            **sel_kwargs: Passed on to ``.sel`` or ``.isel``.

        Returns:
            xr.Dataset: The data of the members from ``cont_names``, combined
                using the given combination method.

        Raises:
            ValueError: On invalid combination method; on conflicting
                coordinate information on group-level and member-level.
            KeyError: In ``concat`` mode, upon missing members.
        """

        def get_cont(
            name: str, combination_method: str
        ) -> Tuple[Union[XrDataContainer, None], str]:
            """Retrieve the container from the group, potentially changing the
            combination method. If no container could be found, returns None,
            which denotes that further processing should be skipped
            """
            try:
                cont = self[name]

            except KeyError as err:
                if combination_method != "concat":
                    # Failing is ok. But cannot do anything else here
                    return None, combination_method

                # Otherwise, should raise!
                raise ItemAccessError(
                    self,
                    key=name,
                    suffix=(
                        "Make sure that the member can be found under this "
                        "name or change the combination method to 'merge' or "
                        "'try_concat'."
                    ),
                ) from err

            else:
                return cont, combination_method

        def process_cont(cont, coords) -> Tuple[xr.DataArray, dict]:
            """Process the given container and coordinates into a data array;
            this applies selection along container dimensions that overlap with
            the group dimensions as well as deep selection.
            """

            def all_equal(x, y) -> bool:
                """Check element-wise equality of two arrays."""
                try:
                    if x.dtype.kind == "f" or y.dtype.kind == "f":
                        return np.allclose(x, y)

                    else:
                        return np.array_equal(x, y)

                except Exception:
                    warnings.warn(
                        "Element-wise array comparison failed. Received: "
                        f"\nx: {x}\ny: {y}\nTreating them as not equal.",
                        DantroWarning,
                    )
                    return False

            # Apply the coordinates of the overlapping dimensions
            # (If there are none, select everything to get a DataArray)
            # Note that all coordinates in `coords` are scalar coordinates
            # since they describe a single entry within the member map.
            overlapping_indexers = {
                dim: coord for dim, coord in coords.items() if dim in cont.dims
            }

            darr = cont.sel(overlapping_indexers, drop=False)
            # The selection above is done with `drop=False`, because `coords`
            # might contain coordinates that were not passed as indexers to
            # `.sel` or `.isel` and therefore should not be dropped.

            # This is to ensure that the array that is used matches only a
            # single coordinate combination, i.e. one _point_ in the space
            # spanned by self.member_map.coords.
            # This operation _might_ increase memory usage temporarily, because
            # data from the same member is accessed multiple times and only one
            # coordinate combination is extracted from it (instead of selecting
            # all at once; which would however require a different architecture
            # and not allow using the convenient xarray interface).

            # Apply the deep indexers
            if deep_indexers:
                if by_index:
                    darr = darr.isel(deep_indexers, **sel_kwargs)
                else:
                    darr = darr.sel(deep_indexers, **sel_kwargs)

            # If the selection is done with `drop=True`, non-dimension
            # coordinates were already removed from `coords`. If there are
            # overlapping dimensions, these have to be applied again.
            dropped_shallow_indexers = {
                dim: coord
                for dim, coord in shallow_indexers.items()
                if dim not in coords and dim in cont.dims
            }

            if dropped_shallow_indexers:
                if by_index:
                    darr = darr.isel(dropped_shallow_indexers, **sel_kwargs)
                else:
                    darr = darr.sel(dropped_shallow_indexers, **sel_kwargs)

            # Finally, check for conflicting non-dimension coordinates in the
            # processed data-array (which were ignored for the selections
            # above).
            confl_coords = {
                d: (darr.coords[d].values, coords[d].values)
                for d in darr.coords
                if d not in darr.dims
                and d in coords
                and not all_equal(darr.coords[d], coords[d])
            }

            if confl_coords:
                # Provide some information on the conflicting coords
                infostr = (
                    "\n"
                    r"Name: {name}"
                    "\n"
                    r"Group-level coordinate: {c_grp}"
                    "\n"
                    r"Member-level coordinate: {c_mem}"
                )

                suffix = "\n".join(
                    [
                        infostr.format(name=k, c_grp=v[1], c_mem=v[0])
                        for k, v in confl_coords.items()
                    ]
                )

                raise ValueError(
                    "Conflicting non-dimension coordinate"
                    f"{'s' if len(confl_coords) > 1 else ''} found within a "
                    f"member of {self.logstr}:\n" + suffix
                )

            return darr, coords

        # If the pre-selected member map contains only a single entry or single
        # invalid entry, no data combination will have to be done, i.e. no
        # `combination_method` needs to be applied. This can speed up the
        # computation time a lot.
        if cont_names.size == 1 or (cont_names != "").sum() == 1:
            if cont_names.size == 1:
                # Select the group member directly.
                cont = self[cont_names.item()]
                # Applying `.squeeze()` here is equivalent to selecting the
                # single entry, since no dimension is of size > 1.
                coords = cont_names.squeeze().coords
            else:
                # Find container name and respective index in `cont_names`
                name = cont_names.where(cont_names != "", drop=True).item()
                idx = tuple(e.item() for e in np.where(cont_names != ""))
                # Extract container and coords
                cont = self[name]
                coords = cont_names[idx].coords

            # Process the container into a DataArray.
            darr, coords = process_cont(cont, coords)

            # Pass on 1D dimensions with the coordinate information
            darr = darr.expand_dims(
                {
                    d: ([c.values] if c.ndim == 0 else c.values)
                    for d, c in coords.items()
                    if d in cont_names.dims
                }
            )

            # Pass on non-dimension coordinates
            darr = darr.assign_coords(
                {
                    d: c.values
                    for d, c in coords.items()
                    if d not in cont_names.dims and d not in darr.coords
                }
            )

            return darr

        dsets = np.zeros(cont_names.shape, dtype="object")
        dsets.fill(dict())  # placeholders, ignored in xr.merge

        # Create an iterator over the container names (mirrors dsets iteration)
        names_iter = np.nditer(cont_names, flags=("multi_index", "refs_ok"))

        for name in names_iter:
            # Get the corresponding member container, potentially changing the
            # combination method
            cont, combination_method = get_cont(
                name.item(), combination_method
            )

            # Might not have been found; go to the next iteration
            if cont is None:
                continue

            # Get the coordinates for this member container and further process
            # the container into a DataArray.
            coords = cont_names[names_iter.multi_index].coords
            darr, coords = process_cont(cont, coords)

            # As it's easier to work on xr.Datasets than on xr.DataArray-like
            # objects, create a dataset from the container, using a temporary
            # name which will later be used to resolve it back to a DataArray
            dset = darr.to_dataset(name="_tmp_dset_name")

            # Now, need to expand the dimensions to accomodate the coordinates.
            # (Important for concatenation!)
            # Add the new dimensions in front and assign coordinates to them.
            dset = dset.expand_dims(
                {
                    d: ([c.values] if c.ndim == 0 else c.values)
                    for d, c in coords.items()
                    if d in cont_names.dims  # only for actual dimensions
                }
            )
            # NOTE While this creates a non-shallow copy of the data, there is
            #      no other way of doing this: a copy can only be avoided if
            #      the DataArray can re-use the existing variables – for the
            #      changes it needs to do to expand the dims, however, it will
            #      necessarily need to create a copy of the original data.
            #      Thus, we might as well let xarray take care of that instead
            #      of bothering with that ourselves ...

            # For coordinates that are not linked to a dimension in
            # `cont_names`, do not expand the dimensions but assign them as
            # non-dimension coordinates.
            dset = dset.assign_coords(
                {
                    d: c.values
                    for d, c in coords.items()
                    if d not in cont_names.dims
                    # If coord already there, don't try assigning it. The check
                    # for conflicting coords ensures that it's correct.
                    and d not in dset.coords
                }
            )

            # Done. Store it in the object-array of datasets
            dsets[names_iter.multi_index] = dset

        # Now ready to combine them.
        if combination_method == "concat":
            dset = self._combine_by_concatenation(dsets, dims=cont_names.dims)

        elif combination_method == "merge":
            dset = self._combine_by_merge(dsets)

        elif combination_method == "try_concat":
            try:
                dset = self._combine_by_concatenation(
                    dsets, dims=cont_names.dims
                )

            except Exception as exc:
                # NOTE The exception is now something other than a member
                #      missing, i.e. some numerical issue during concatenation
                # Try again with merging ...
                log.warning(
                    "Failed concatenation with %s: %s. Attempting merge ...",
                    exc.__class__.__name__,
                    exc,
                )
                dset = self._combine_by_merge(dsets)

        else:
            raise ValueError(
                f"Invalid combination_method argument: {combination_method}! "
                "Available methods: try_concat, concat, merge."
            )

        # Combined into one dataset now, with '_tmp_dset_name' data variable...
        # Convert back into a DataArray; can drop the temporary name now.
        darr = dset["_tmp_dset_name"]
        darr.name = None
        return darr

    @classmethod
    def _combine_by_merge(cls, dsets: np.ndarray) -> xr.Dataset:
        """Combine the given datasets by merging using `xarray.merge`.

        Args:
            dsets (np.ndarray): The object-dtype array of xr.Datasets that are
                to be combined.

        Returns:
            xr.Dataset: All datasets, aligned and combined via `xarray.merge`
        """
        log.debug("Combining %d datasets by merging ...", dsets.size)

        dset = xr.merge(dsets.flat)

        log.debug("Merge successful.")
        return dset

    @classmethod
    def _combine_by_concatenation(
        cls, dsets: np.ndarray, *, dims: TDims
    ) -> xr.Dataset:
        """Combine the given datasets by concatenation using `xarray.concat`
        and subsequent application along all dimensions specified in ``dims``.

        Args:
            dsets (np.ndarray): The object-dtype array of xr.Dataset objects
                that are to be combined by concatenation.
            dims (TDims): The dimension names corresponding to _all_ the
                dimensions of the ``dsets`` array.

        Returns:
            xr.Dataset: The dataset resulting from the concatenation
        """
        log.debug(
            "Combining %d datasets by concatenation along %d dimension%s ...",
            dsets.size,
            len(dsets.shape),
            "s" if len(dsets.shape) != 1 else "",
        )

        # Go over all dimensions and concatenate
        # This effectively reduces the dsets array by one dimension in each
        # iteration by applying the xr.concat function along the axis
        # NOTE np.apply_along_axis would be what is desired here, but that
        #      function unfortunately tries to cast objects to np.arrays
        #      which is not what we want here at all!
        #      Thus, there is one implemented in dantro.tools ...
        for dim_idx, dim_name in reversed(list(enumerate(dims))):
            log.debug(
                "Concatenating along axis '%s' (axis # %d) ...",
                dim_name,
                dim_idx,
            )

            dsets = apply_along_axis(
                xr.concat, axis=dim_idx, arr=dsets, dim=dim_name
            )

        log.debug("Concatenation successful.")

        # The single item in the now scalar array is the combined xr.Dataset
        return dsets.item()



"""Plotting functions that can be used by the
:py:class:`~dantro.plot_creators.pcr_ext.ExternalPlotCreator` and derived plot
creators.
"""

from .basic import lineplot
from .generic import errorbands, errorbar, facet_grid
from .multiplot import multiplot



"""Supplies loading functions for YAML files"""

import logging

from ..containers import MutableMappingContainer, ObjectContainer
from ..tools import load_yml
from ._tools import add_loader

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class YamlLoaderMixin:
    """Supplies functionality to load yaml files in the data manager"""

    @add_loader(TargetCls=MutableMappingContainer)
    def _load_yaml(
        filepath: str, *, TargetCls: type
    ) -> MutableMappingContainer:
        """Load a yaml file from the given path and creates a container to
        store that data in.

        Args:
            filepath (str): Where to load the yaml file from
            TargetCls (type): The class constructor

        Returns:
            MutableMappingContainer: The loaded yaml file as a container
        """
        # Load the dict
        d = load_yml(filepath)

        # Populate the target container with the data
        return TargetCls(data=d, attrs=dict(filepath=filepath))

    @add_loader(TargetCls=ObjectContainer)
    def _load_yaml_to_object(
        filepath: str, *, TargetCls: type
    ) -> ObjectContainer:
        """Load a yaml file from the given path and creates a container to
        store that data in.

        Args:
            filepath (str): Where to load the yaml file from
            TargetCls (type): The class constructor

        Returns:
            ObjectContainer: The loaded yaml file as a container
        """
        # Load the dict
        d = load_yml(filepath)

        # Populate the target container with the data
        return TargetCls(data=d, attrs=dict(filepath=filepath))

    # Also make available under `yml`
    _load_yml = _load_yaml
    _load_yml_to_object = _load_yaml_to_object



"""In this module, the BaseDataGroup is specialized for holding members in a
specific order.
"""

import collections
import logging
from typing import Generator, List

from ..base import BaseDataGroup
from ..mixins import IntegerItemAccessMixin
from ..utils import IntOrderedDict

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class OrderedDataGroup(BaseDataGroup, collections.abc.MutableMapping):
    """The OrderedDataGroup class manages groups of data containers, preserving
    the order in which they were added to this group.

    It uses an OrderedDict to associate containers with this group.
    """

    # Use OrderedDict for storage in insertion order
    _STORAGE_CLS = collections.OrderedDict


# -----------------------------------------------------------------------------


class IndexedDataGroup(IntegerItemAccessMixin, OrderedDataGroup):
    """The IndexedDataGroup class holds members that are of the same type and
    have names that can directly be interpreted as positive integers.

    Especially, this group maintains the correct order of members according to
    integer ordering.

    To speed up element insertion, this group keeps track of recently added
    container names, which are then used as hints for subsequent insertions.

    .. note::

        Albeit the members of this group being ordered, item access still
        refers to the *names* of the members, not their index within the
        sequence!

    .. warning::

        With the underlying ordering mechanism of
        :py:class:`~dantro.utils.ordereddict.KeyOrderedDict`, the performance
        of this data structure is sensitive to the insertion order of elements.

        It is fastest for **in-order** insertions, where the complexity per
        insertion is constant (regardless of whether insertion order is
        ascending or descending).
        For **out-of-order** insertions, the whole key map may have to be
        searched, in which case the complexity scales with the number of
        elements in this group.

    .. hint::

        If experiencing trouble with the performance of this data structure,
        **sort elements before adding them to this group**.
    """

    # A dict of (key length -> last key inserted of that length), which is used
    # as an insertion hint when adding a container to this group
    __last_keys = None

    # Use an orderable dict for storage, i.e. something like OrderedDict, but
    # where it's not sorted by insertion order but by key.
    _STORAGE_CLS = IntOrderedDict

    # The child class should not necessarily be of the same type as this class.
    _NEW_GROUP_CLS = OrderedDataGroup

    # Advanced key access .....................................................

    def key_at_idx(self, idx: int) -> str:
        """Get a key by its index within the container. Can be negative.

        Args:
            idx (int): The index within the member sequence

        Returns:
            str: The desired key

        Raises:
            IndexError: Index out of range
        """
        # Imitate indexing behaviour of lists, tuples, ...
        if not isinstance(idx, int):
            raise TypeError(
                "Expected integer, got {} '{}'!".format(type(idx), idx)
            )

        if idx >= len(self) or idx < -len(self):
            raise IndexError(
                "Index {:d} out of range for {} with {} members!".format(
                    idx, self.logstr, len(self)
                )
            )

        # Wraparound negative
        idx = idx if idx >= 0 else idx % len(self)

        for i, k in enumerate(self.keys()):
            if i == idx:
                return k

    def keys_as_int(self) -> Generator[int, None, None]:
        """Returns an iterator over keys as integer values"""
        for k in self.keys():
            yield int(k)

    # Customizations of parent methods ........................................

    def _add_container_to_data(self, cont) -> None:
        """Adds a container to the underlying integer-ordered dictionary.

        Unlike the parent method, this uses
        :py:meth:`~dantro.utils.ordereddict.KeyOrderedDict.insert` in order to
        provide hints regarding the insertion position. It is optimised for
        insertion in *ascending* order.
        """
        # Keep track of insertion hints
        if self.__last_keys is None:
            self.__last_keys = dict()

        # Insert it, using hint information for names of this length
        self._data.insert(
            cont.name, cont, hint_after=self.__last_keys.get(len(cont.name))
        )

        # Update the hints for names of this length
        self.__last_keys[len(cont.name)] = cont.name

    def _ipython_key_completions_(self) -> List[int]:
        """For ipython integration, return a list of available keys.

        Unlike the BaseDataGroup method, which returns a list of strings, this
        returns a list of integers.
        """
        return list(self.keys_as_int())



"""This sub-package implements non-abstract dantro plot creator classes"""

from .pcr_base import BasePlotCreator, SkipPlot
from .pcr_decl import DeclarativePlotCreator
from .pcr_ext import ExternalPlotCreator, PlotHelper, is_plot_func
from .pcr_psp import MultiversePlotCreator, UniversePlotCreator
from .pcr_vega import VegaPlotCreator

# And gather them into a dictionary that gives names to each of them
ALL = dict(
    external=ExternalPlotCreator,
    declarative=DeclarativePlotCreator,
    universe=UniversePlotCreator,
    multiverse=MultiversePlotCreator,
    vega=VegaPlotCreator,
)



"""The groups sub-package implements BaseDataGroup specializations

isort:skip_file
"""

from .ordered import IndexedDataGroup, OrderedDataGroup
from .graph import GraphGroup
from .labelled import LabelledDataGroup
from .pspgrp import ParamSpaceGroup, ParamSpaceStateGroup
from .time_series import HeterogeneousTimeSeriesGroup, TimeSeriesGroup



#!/usr/bin/env python3
# fmt: off

from setuptools import find_packages, setup

# Dependencies for dantro itself
install_deps = [
    "numpy>=1.19.4",
    "xarray>=0.16.2",
    "dask>=2.10.1",
    "toolz>=0.10.0",        # For dask.delayed
    "distributed>=2.10.0",  # For dask's distributed scheduler
    "scipy>=1.5.4",         # Used as a netcdf4 storage engine for xarray
                            # NOTE With scipy 1.6, will lose py36 support
    "sympy>=1.6.1",
    "h5py>=3.1",
    "matplotlib>=3.2.1",
    "seaborn>=0.11",
    "networkx>=2.5",
    "ruamel.yaml>=0.16.12",
    "dill>=0.3.3",          # For faster and more powerful pickling
    "paramspace>=2.5.5",
]
# NOTE When changing any of the dependencies, make sure to update the table of
#      dependencies in README.md.
#      When adding a NEW dependency, make sure to denote it in the isort
#      configuration, see pyproject.toml.


# Minimal versions of all of the above
minimal_install_deps = [dep.replace(">=", "==") for dep in install_deps]

# Dependencies for running tests and general development of dantro
test_deps = [
    "pytest>=3.4.0",
    "pytest-cov>=2.5.1",
    "tox>=3.1.2",
    "pre-commit>=2.8.2",
]

# Dependencies for building the dantro documentation
doc_deps = [
    "sphinx>=2.4,<3.0",
    "sphinx_rtd_theme>=0.5",
    "ipython>=7.0",
]

# .............................................................................

DESCRIPTION = "Handle, transform, and visualize hierarchically structured data"
LONG_DESCRIPTION = """
``dantro``: handle, transform, and visualize hierarchically structured data
===========================================================================

``dantro`` – from *data* and *dentro* (Greek for *tree*) – is a Python
package that provides a uniform interface for hierarchically structured
and semantically heterogeneous data. It is built around three main
features:

-  **data handling:** loading heterogeneous data into a tree-like data
   structure and providing a uniform interface for it
-  **data transformation:** performing arbitrary operations on the data,
   if necessary using lazy evaluation
-  **data visualization:** creating a visual representation of the
   processed data

Together, these stages constitute a **data processing pipeline**: an
automated sequence of predefined, configurable operations. Akin to a
Continuous Integration pipeline, a data processing pipeline provides a
uniform, consistent, and easily extensible infrastructure that
contributes to more efficient and reproducible workflows. This can be
beneficial especially in a scientific context, for instance when
handling data that was generated by computer simulations.

``dantro`` is meant to be *integrated* into projects and be used to set up
such a data processing pipeline, customized to the needs of the project.
It is designed to be **easily customizable** to the requirements of the project
it is integrated in, even if the involved data is hierachically structured or
semantically heterogeneous.
Furthermore, it allows a **configuration-based specification** of all
operations via YAML configuration files; the resulting pipeline can then be
controlled entirely via these configuration files and without requiring
code changes.

The ``dantro`` package is **open source software** released under the
`LGPLv3+ <(https://www.gnu.org/licenses/lgpl-3.0.html>`_ license.
It was developed alongside the `Utopia project <https://ts-gitlab.iup.uni-heidelberg.de/utopia/utopia>`_,
but is an independent package.

Learn more
----------

* `Documentation <https://dantro.readthedocs.io/>`_
* `Project page <https://ts-gitlab.iup.uni-heidelberg.de/utopia/dantro>`_
* `README and installation instructions <https://ts-gitlab.iup.uni-heidelberg.de/utopia/dantro#installing-dantro>`_
* `Publication in the Journal of Open Source Software <https://doi.org/10.21105/joss.02316>`_

"""


# .............................................................................
# fmt: on

# A function to extract version number from __init__.py
def find_version(*file_paths) -> str:
    """Tries to extract a version from the given path sequence"""
    import codecs
    import os
    import re

    def read(*parts):
        """Reads a file from the given path sequence, relative to this file"""
        here = os.path.abspath(os.path.dirname(__file__))
        with codecs.open(os.path.join(here, *parts), "r") as fp:
            return fp.read()

    # Read the file and match the __version__ string
    file = read(*file_paths)
    match = re.search(r"^__version__\s?=\s?['\"]([^'\"]*)['\"]", file, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string in " + str(file_paths))


# .............................................................................

setup(
    name="dantro",
    #
    # Set the version from dantro.__version__
    version=find_version("dantro", "__init__.py"),
    #
    # Project info
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author="dantro developers",
    author_email="dantro-dev@iup.uni.heidelberg.de",
    license="LGPL-3.0-or-later",
    url="https://ts-gitlab.iup.uni-heidelberg.de/utopia/dantro",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or"
        " later (LGPLv3+)",
    ],
    #
    # Distribution details, dependencies, ...
    packages=find_packages(exclude=["tests.*", "tests"]),
    data_files=[("", ["COPYING", "COPYING.LESSER", "README.md"])],
    install_requires=install_deps,
    tests_require=test_deps,
    test_suite="py.test",
    extras_require=dict(
        minimal_deps=minimal_install_deps,
        test=test_deps,
        doc=doc_deps,
        dev=test_deps + doc_deps,
    ),
)



"""Definition of an OrderedDict-subclass that maintains the order key values
rather than the insertion order. It can be specialized to use a specific
comparison function for ordering keys.
"""

import collections
import logging

# Imports needed for replicating OrderedDict behaviour
import sys as _sys
from collections import (
    _Link,
    _OrderedDictItemsView,
    _OrderedDictKeysView,
    _OrderedDictValuesView,
    _proxy,
)
from collections.abc import MutableMapping
from operator import eq as _eq
from reprlib import recursive_repr as _recursive_repr
from typing import Callable

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class KeyOrderedDict(dict):
    """This dict maintains the order of keys not by their insertion but by
    their value. It is a re-implementation of collections.OrderedDict, because
    subclassing that class is really difficult.

    Ordering is maintained by adjusting those methods of OrderedDict that
    take care of building and maintaining the doubly-linked list that provides
    the ordering. See OrderedDict for details, e.g. the weak-referencing.

    In effect, this relates only to ``__setitem__``; all other methods rely on
    this to add elements to the mapping.

    For comparison, the ``key`` callable given at initialisation can be used to
    perform a operation on keys, the result of which is used in comparison.
    If this is not given, the ``DEFAULT_KEY_COMPARATOR`` class variable is
    used; note that this needs to be a binary function, where the first
    argument is equivalent to ``self`` and the second is the actual key to
    perform the unary operation on.
    """

    # A counter to keep track of the total number of comparisons made during
    # element ordering, e.g. when adding elements to this dict.
    __num_comparisons = 0

    # The default key comparator
    DEFAULT_KEY_COMPARATOR = lambda _, k: k

    def __init__(self, *args, key: Callable = None, **kwds):
        """Initialize a KeyOrderedDict, which maintains key ordering. If no
        custom ordering function is given, orders by simple "smaller than"
        comparison.

        Apart from that, the interface is the same as for regular dictionaries.

        Args:
            *args: a single sequence of (key, value) pairs to insert
            key (Callable, optional): The callable used to
            **kwds: Passed on to ``update`` method

        Raises:
            TypeError: on len(args) > 1
        """
        if len(args) > 1:
            raise TypeError(
                "expected at most 1 arguments, got {}".format(len(args))
            )

        # Set the key comparison function, passing through if nothing is set
        self._key = key if key is not None else self.DEFAULT_KEY_COMPARATOR

        # Set up internally used attributes
        try:
            self.__root

        except AttributeError:
            self.__hardroot = _Link()
            self.__root = root = _proxy(self.__hardroot)
            root.prev = root.next = root
            self.__map = {}

        # Populate the dict
        self.__update(*args, **kwds)

    # .........................................................................
    # Custom implementations that ascertain key ordering

    def _key_comp_lt(self, k1, k2) -> bool:
        """The key comparator. Returns true for k1 < k2.

        Before comparison, the unary ``self._key`` method is invoked on both
        keys.

        Args:
            k1: lhs of comparison
            k2: rhs of comparison

        Returns:
            bool: k1 < k2
        """
        # Create the actual keys to compare
        try:
            _k1, _k2 = self._key(k1), self._key(k2)

        except Exception as exc:
            raise ValueError(
                "Could not apply key transformation method on "
                f"one or both of the keys '{k1}' and '{k2}'!"
            ) from exc

        # Increment the comparisons counter
        self.__num_comparisons += 1

        # Compare
        try:
            return _k1 < _k2

        except Exception as exc:
            raise ValueError(
                f"Failed comparing '{_k1}' (type {type(_k1)}, from key '{k1}')"
                f" to '{_k2}' (type {type(_k2)}, from key '{k2}')! "
                f"Keys of {self.__class__.__name__} need to be comparable "
                f"after the key transformation function {self._key.__name__} "
                "was applied to them."
            ) from exc

    def __setitem__(
        self,
        key,
        value,
        *,
        dict_setitem=dict.__setitem__,
        proxy=_proxy,
        Link=_Link,
        start=None,
    ):
        """Set the item with the provided key to the given value, maintaining
        the ordering specified by ``_key_comp_lt``.

        If the key is not available, this takes care of finding the right place
        to insert the new link in the doubly-linked list.

        Unlike the regular ``__setitem__``, this allows specifying a start
        element at which to begin the search for an insertion point, which may
        greatly speed up insertion. By default, it is attempted to insert
        after the last element; if that is not possible, a full search is done.

        .. warning::

            This operation does not scale well with out-of-order insertion!

            This behavior is inherent to this data structure, where key
            ordering has to be maintained during every insertion. While a best
            guess is made regarding the insertion points (see above), inserting
            elements completely out of order will require a search time that is
            proportional to the number of elements *for each insertion*.

        .. hint::

            If you have information about where the element should be stored,
            use :py:meth:`~dantro.utils.ordereddict.KeyOrderedDict.insert` and
            provide the ``hint_after`` argument.
        """
        if key not in self:  # complexity: O(1)
            # Create a new link in the inherited mapping
            self.__map[key] = link = Link()

            # Find the link in the doubly-linked list after which the new
            # element is to be inserted, i.e.: the first element that compares
            # true when invoking the key comparator
            element = self.__find_element_to_insert_after(key, start=start)

            # element is now the element _after_ which to insert the new link
            link.prev, link.next, link.key = element, element.next, key

            # Update neighbouring links such that the new link is in between
            link.prev.next = link
            link.next.prev = proxy(link)  # need be a weak link

        # Key is now available
        # Set the value of the item using the method implemented by dict
        dict_setitem(self, key, value)

    def insert(self, key, value, *, hint_after=None):
        """Inserts a (key, value) pair using hint information to speed up the
        search for an insertion point.

        If hint information is available, it is highly beneficial to add this

        Args:
            key: The key at which to insert
            value: The value to insert
            hint_after (optional): A best guess after which key to insert.
                The requirement here is that the key compares
        """
        if hint_after is not None:
            self.__setitem__(key, value, start=self.__map.get(hint_after))

        else:
            # No hints, use regular insertion
            self[key] = value

    def __find_element_to_insert_after(self, key, *, start=None) -> _Link:
        """Finds the link in the doubly-linked list after which a new element
        with the given key may be inserted, i.e. the last element that
        compares False when invoking the key comparator.

        If inserting an element to the back or the front of the key list, the
        complexity of this method is constant. Otherwise, it scales with the
        number of already existing elements.

        Args:
            key: The key to find the insertion spot for
            start (None, optional): A key to use to start looking, if not given
                will use the last element as a best guess.
        """
        log.trace("Looking for insertion point for key '%s' ...", key)

        # Set default value: the last element; this is a best guess which is
        # very successful if inserting in ascending order and not very costly
        # (one additional comparison) if inserting in descending order.
        # This allows constant insertion time for both these scenarios.
        if start is None:
            start = self.__root.prev

        # For unpopulated maps, can directly insert at this point
        if start.next is start:
            log.trace("  Map is empty; inserting in the beginning.")
            return start

        # For populated maps, need to check whether the start element is really
        # suitable:
        #   * The start element needs to compare false in key < start.key,
        #     otherwise we don't know whether the next "stricter greater than"
        #     comparison is really the first (which is the requirement for a
        #     valid starting point).
        #   * Also, we can't do a key comparison for the root element, in which
        #     case we can also use this fallback case.
        if start is self.__root or self._key_comp_lt(key, start.key):
            log.trace("  Start element not suitable; using sentinel instead.")
            start = self.__root

        # Now begin element search at the specified start element
        element = start

        # Otherwise, need to iterate over the linked list until back at root
        while element.next is not self.__root:
            # Check if the next element's key would compare strictly
            # greater than the key to be inserted, in which case we found the
            # element to insert after
            if self._key_comp_lt(key, element.next.key):
                break

            # Did not find it. Go to the next one
            element = element.next

        log.trace(
            "  Found insertion point: after '%s'. Total comparisons: %d",
            element.key if element is not self.__root else "<root>",
            self._num_comparisons,
        )
        return element

    @property
    def _num_comparisons(self) -> int:
        """Total number of comparisons performed between elements, e.g. when
        finding an insertion point

        This number is low, if the insertion order is sequential.
        For out-of-order insertion, it may become large.
        """
        return self.__num_comparisons

    # .........................................................................
    # Remaining dict interface, oriented at collections.OrderedDict

    def __delitem__(self, key, dict_delitem=dict.__delitem__):
        """kod.__delitem__(y) <==> del kod[y]

        Deleting an existing item uses self.__map to find the link which gets
        removed by updating the links in the predecessor and successor nodes.
        """
        dict_delitem(self, key)
        link = self.__map.pop(key)
        link_prev = link.prev
        link_next = link.next
        link_prev.next = link_next
        link_next.prev = link_prev
        link.prev = None
        link.next = None

    def __iter__(self):
        """kod.__iter__() <==> iter(kod)

        Traverse the linked list in order.
        """
        root = self.__root
        curr = root.next
        while curr is not root:
            yield curr.key
            curr = curr.next

    def __reversed__(self):
        """kod.__reversed__() <==> reversed(kod)

        Traverse the linked list in reverse order.
        """
        root = self.__root
        curr = root.prev
        while curr is not root:
            yield curr.key
            curr = curr.prev

    def clear(self):
        """Remove all items"""
        root = self.__root
        root.prev = root.next = root
        self.__map.clear()
        dict.clear(self)

    def __sizeof__(self):
        """Get the size of this object"""
        sizeof = _sys.getsizeof

        # number of links including root
        n = len(self) + 1

        # instance dictionary
        size = sizeof(self.__dict__)

        # internal dict and inherited dict
        size += sizeof(self.__map) * 2

        # link and proxy objects
        size += sizeof(self.__hardroot) * n
        size += sizeof(self.__root) * n

        # key comparator
        size += sizeof(self._key)

        return size

    update = __update = MutableMapping.update

    def keys(self):
        """D.keys() -> a set-like object providing a view on D's keys"""
        return _OrderedDictKeysView(self)

    def items(self):
        """D.items() -> a set-like object providing a view on D's items"""
        return _OrderedDictItemsView(self)

    def values(self):
        """D.values() -> an object providing a view on D's values"""
        return _OrderedDictValuesView(self)

    __ne__ = MutableMapping.__ne__

    __marker = object()

    def pop(self, key, default=__marker):
        """kod.pop(k[,d]) -> v, remove specified key and return the
        corresponding value.  If key is not found, default is returned if
        given, otherwise KeyError is raised.
        """
        if key in self:
            result = self[key]
            del self[key]
            return result
        if default is self.__marker:
            raise KeyError(key)
        return default

    def setdefault(self, key, default=None):
        """kod.setdefault(k[,d]) -> kod.get(k,d), also set kod[k]=d if k not
        in kod
        """
        if key in self:
            return self[key]
        self[key] = default
        return default

    @_recursive_repr()
    def __repr__(self):
        """kod.__repr__() <==> repr(kod)"""
        if not self:
            return "%s()" % (self.__class__.__name__,)
        return "%s(%r)" % (self.__class__.__name__, list(self.items()))

    def __reduce__(self):
        """Return state information for pickling"""
        inst_dict = vars(self).copy()
        for k in vars(KeyOrderedDict()):
            inst_dict.pop(k, None)
        return self.__class__, (), inst_dict or None, None, iter(self.items())

    def copy(self):
        """kod.copy() -> a shallow copy of kod, maintaining the key comparator"""
        return self.__class__(self, key=self._key)

    @classmethod
    def fromkeys(cls, iterable, value=None, *, key: Callable = None):
        """KOD.fromkeys(S[, v]) -> New key-ordered dictionary with keys from S.
        If not specified, the value defaults to None.
        """
        self = cls(key=key)
        for key in iterable:
            self[key] = value
        return self

    def __eq__(self, other):
        """kod.__eq__(y) <==> kod==y.
        Comparison to another KeyOrderedDict or OrderedDict is order-sensitive
        while comparison to a regular mapping is order-insensitive.
        """
        if isinstance(other, (KeyOrderedDict, collections.OrderedDict)):
            return dict.__eq__(self, other) and all(map(_eq, self, other))
        return dict.__eq__(self, other)


# -----------------------------------------------------------------------------
# Specializations


class IntOrderedDict(KeyOrderedDict):
    """A :py:class:`~dantro.utils.ordereddict.KeyOrderedDict` specialization
    that assumes keys to be castable to integer and using the comparison of
    the resulting integer values for maintaining the order
    """

    DEFAULT_KEY_COMPARATOR = lambda _, k: int(k)



"""This module implements the DataManager class, the root of the data tree."""

import copy
import datetime
import glob
import logging
import os
import re
import time
import warnings
from typing import Callable, Dict, List, Tuple, Union

import dill as pkl

from ._hash import _hash
from .base import PATH_JOIN_CHAR, BaseDataContainer, BaseDataGroup
from .exceptions import *
from .groups import OrderedDataGroup
from .tools import (
    clear_line,
    fill_line,
    format_bytesize,
    load_yml,
    recursive_update,
)

# Local constants
log = logging.getLogger(__name__)

DATA_TREE_DUMP_EXT = ".d3"


# -----------------------------------------------------------------------------


class DataManager(OrderedDataGroup):
    """The DataManager is the root of a data tree, coupled to a specific data
    directory.

    It handles the loading of data and can be used for interactive work with
    the data.
    """

    # Use a base load configuration to start with
    _BASE_LOAD_CFG = None

    # Define some default groups (same syntax create_groups argument)
    _DEFAULT_GROUPS = None

    # Define as class variable what should be the default group type
    _DATA_GROUP_DEFAULT_CLS = OrderedDataGroup

    # For simple lookups, store class names in a dict; not set by default
    _DATA_GROUP_CLASSES = None

    # The default tree file cache path, parsed by _parse_file_path
    _DEFAULT_TREE_CACHE_PATH = ".tree_cache.d3"

    # .........................................................................
    # Initialization

    def __init__(
        self,
        data_dir: str,
        *,
        name: str = None,
        load_cfg: Union[dict, str] = None,
        out_dir: Union[str, bool] = "_output/{timestamp:}",
        out_dir_kwargs: dict = None,
        create_groups: List[Union[str, dict]] = None,
        condensed_tree_params: dict = None,
        default_tree_cache_path: str = None,
    ):
        """Initializes a DataManager for the specified data directory.

        Args:
            data_dir (str): the directory the data can be found in. If this is
                a relative path, it is considered relative to the current
                working directory.
            name (str, optional): which name to give to the DataManager. If no
                name is given, the data directories basename will be used
            load_cfg (Union[dict, str], optional): The base configuration used
                for loading data. If a string is given, assumes it to be the
                path to a YAML file and loads it using the
                :py:func:`~dantro._yaml.load_yml` function. If None is given,
                it can still be supplied to the
                :py:meth:`~dantro.data_mngr.DataManager.load` method later on.
            out_dir (Union[str, bool], optional): where output is written to.
                If this is given as a relative path, it is considered relative
                to the ``data_dir``. A formatting operation with the keys
                ``timestamp`` and ``name`` is performed on this, where the
                latter is the name of the data manager. If set to False, no
                output directory is created.
            out_dir_kwargs (dict, optional): Additional arguments that affect
                how the output directory is created.
            create_groups (List[Union[str, dict]], optional): If given, these
                groups will be created after initialization. If the list
                entries are strings, the default group class will be used; if
                they are dicts, the `name` key specifies the name of the group
                and the `Cls` key specifies the type. If a string is given
                instead of a type, the lookup happens from the
                ``_DATA_GROUP_CLASSES`` variable.
            condensed_tree_params (dict, optional): If given, will set the
                parameters used for the condensed tree representation.
                Available options: ``max_level`` and ``condense_thresh``, where
                the latter may be a callable.
                See :py:meth:`dantro.base.BaseDataGroup._tree_repr` for more
                information.
            default_tree_cache_path (str, optional): The path to the default
                tree cache file. If not given, uses the value from the class
                variable ``_DEFAULT_TREE_CACHE_PATH``. Whichever value was
                chosen is then prepared using the
                :py:meth:`~dantro.data_mngr.DataManager._parse_file_path`
                method, which regards relative paths as being relative to the
                associated data directory.
        """
        # Find a name if none was given
        if not name:
            basename = os.path.basename(os.path.abspath(data_dir))
            name = "{}_Manager".format(basename.replace(" ", "_"))

        log.info("Initializing %s '%s'...", self.classname, name)

        # Initialize as a data group via parent class
        super().__init__(name=name)

        # Set condensed tree parameters
        if condensed_tree_params:
            self._set_condensed_tree_params(**condensed_tree_params)

        # Initialize directories
        self.dirs = self._init_dirs(
            data_dir=data_dir,
            out_dir=out_dir,
            **(out_dir_kwargs if out_dir_kwargs else {}),
        )

        # Parse the default tree cache path
        _tcp = (
            default_tree_cache_path
            if default_tree_cache_path
            else self._DEFAULT_TREE_CACHE_PATH
        )
        self._tree_cache_path = self._parse_file_path(_tcp)

        # Start out with the default load configuration or, if not given, with
        # an empty one
        self.load_cfg = {} if not self._BASE_LOAD_CFG else self._BASE_LOAD_CFG

        # Resolve string arguments
        if isinstance(load_cfg, str):
            # Assume this is the path to a configuration file and load it
            log.debug(
                "Loading the default load config from a path:\n  %s", load_cfg
            )
            load_cfg = load_yml(load_cfg)

        # If given, use it to recursively update the base
        if load_cfg:
            self.load_cfg = recursive_update(self.load_cfg, load_cfg)

        # Create default groups, as specified in the _DEFAULT_GROUPS class
        # variable and the create_groups argument
        if self._DEFAULT_GROUPS or create_groups:
            # Parse both into a new list to iterate over
            specs = self._DEFAULT_GROUPS if self._DEFAULT_GROUPS else []
            if create_groups:
                specs += create_groups

            log.debug(
                "Creating %d empty groups from defaults and/or given "
                "initialization arguments ...",
                len(specs),
            )
            for spec in specs:
                if isinstance(spec, dict):
                    # Got a more elaborate group specification
                    self.new_group(**spec)

                else:
                    # Assume this is the group name; use the default class
                    self.new_group(spec)

        # Done
        log.debug("%s initialized.", self.logstr)

    def _set_condensed_tree_params(self, **params):
        """Helper method to set the ``_COND_TREE_*`` class variables"""
        available_keys = ("max_level", "condense_thresh")

        for key, value in params.items():
            if key.lower() not in available_keys:
                raise KeyError(
                    "Invalid condensed tree parameter: '{}'! The "
                    "available keys are: {}."
                    "".format(key, ", ".join(available_keys))
                )
            setattr(self, "_COND_TREE_" + key.upper(), value)

    def _init_dirs(
        self,
        *,
        data_dir: str,
        out_dir: Union[str, bool],
        timestamp: float = None,
        timefstr: str = "%y%m%d-%H%M%S",
        exist_ok: bool = False,
    ) -> Dict[str, str]:
        """Initializes the directories managed by this DataManager and returns
        a dictionary that stores the absolute paths to these directories.

        If they do not exist, they will be created.

        Args:
            data_dir (str): the directory the data can be found in. If this is
                a relative path, it is considered relative to the current
                working directory.
            out_dir (Union[str, bool]): where output is written to.
                If this is given as a relative path, it is considered relative
                to the **data directory**. A formatting operation with the
                keys ``timestamp`` and ``name`` is performed on this, where
                the latter is the name of the data manager. If set to False,
                no output directory is created.
            timestamp (float, optional): If given, use this time to generate
                the `date` format string key. If not, uses the current time.
            timefstr (str, optional): Format string to use for generating the
                string representation of the current timestamp
            exist_ok (bool, optional): Whether the output directory may exist.
                Note that it only makes sense to set this to True if you can
                be sure that there will be no file conflicts! Otherwise the
                errors will just occur at a later stage.

        Returns:
            Dict[str, str]: The directory paths registered under certain keys,
                e.g. ``data`` and ``out``.
        """

        # Make the data directory absolute
        log.debug("Received `data_dir` argument:\n  %s", data_dir)
        data_dir = os.path.abspath(data_dir)

        # Save dictionary that will hold info on directories
        dirs = dict(data=data_dir)

        # See if an output directory should be created
        if out_dir:
            log.debug("Received `out_dir` argument:\n  %s", out_dir)

            # Make current date and time available for formatting operations
            time = (
                datetime.datetime.fromtimestamp(timestamp)
                if timestamp
                else datetime.datetime.now()
            )
            timestr = time.strftime(timefstr)

            # Perform a format operation on the output directory
            out_dir = out_dir.format(name=self.name.lower(), timestamp=timestr)

            # If it is relative, assume it to be relative to the data directory
            if not os.path.isabs(out_dir):
                # By joining them together, out_dir is now relative
                out_dir = os.path.join(data_dir, out_dir)

            # Make path absolute and store in dict
            dirs["out"] = os.path.abspath(out_dir)

            # Create the directory
            os.makedirs(dirs["out"], exist_ok=exist_ok)

        else:
            dirs["out"] = False

        # Inform about the managed directories, then return
        log.debug(
            "Managed directories:\n%s",
            "\n".join(["  {:>8s} : {}".format(k, v) for k, v in dirs.items()]),
        )

        return dirs

    @property
    def hashstr(self) -> str:
        """The hash of a DataManager is computed from its name and the coupled
        data directory, which are regarded as the relevant parts. While other
        parts of the DataManager are not invariant, it is characterized most by
        the directory it is associated with.

        As this is a string-based hash, it is not implemented as the __hash__
        magic method but as a separate property.

        WARNING Changing how the hash is computed for the DataManager will
                invalidate all TransformationDAG caches.
        """
        return _hash(
            "<DataManager '{}' @ {}>".format(self.name, self.dirs["data"])
        )

    def __hash__(self) -> int:
        """The hash of this DataManager, computed from the hashstr property"""
        return hash(self.hashstr)

    @property
    def tree_cache_path(self) -> str:
        """Absolute path to the default tree cache file"""
        return self._tree_cache_path

    @property
    def tree_cache_exists(self) -> bool:
        """Whether the tree cache file exists"""
        return os.path.isfile(self._tree_cache_path)

    # .........................................................................
    # Loading data

    def load_from_cfg(
        self,
        *,
        load_cfg: dict = None,
        update_load_cfg: dict = None,
        exists_action: str = "raise",
        print_tree: Union[bool, str] = False,
    ) -> None:
        """Load multiple data entries using the specified load configuration.

        Args:
            load_cfg (dict, optional): The load configuration to use. If not
                given, the one specified during initialization is used.
            update_load_cfg (dict, optional): If given, it is used to update
                the load configuration recursively
            exists_action (str, optional): The behaviour upon existing data.
                Can be: ``raise`` (default), ``skip``, ``skip_nowarn``,
                ``overwrite``, ``overwrite_nowarn``. With the ``*_nowarn``
                values, no warning is given if an entry already existed.
            print_tree (Union[bool, str], optional): If True, the full tree
                representation of the DataManager is printed after the data
                was loaded. If ``'condensed'``, the condensed tree will be
                printed.

        Raises:
            TypeError: Raised if a given configuration entry was of invalid
                type, i.e. not a dict
        """
        # Determine which load configuration to use
        if not load_cfg:
            log.debug(
                "No new load configuration given; will use load "
                "configuration given at initialization."
            )
            load_cfg = self.load_cfg

        # Make sure to work on a copy, be it on the defaults or on the passed
        load_cfg = copy.deepcopy(load_cfg)

        if update_load_cfg:
            # Recursively update with the given keywords
            load_cfg = recursive_update(load_cfg, update_load_cfg)
            log.debug("Updated the load configuration.")

        log.hilight("Loading %d data entries ...", len(load_cfg))

        # Loop over the data entries that were configured to be loaded
        for entry_name, params in load_cfg.items():
            # Check if this is of valid type
            if not isinstance(params, dict):
                raise TypeError(
                    "Got invalid load specifications for entry "
                    f"'{entry_name}'! Expected dict, got {type(params)} with "
                    f"value '{params}'. Check the correctness of the given "
                    "load configuration!"
                )

            # Use the public method to load this single entry
            self.load(
                entry_name,
                exists_action=exists_action,
                print_tree=False,  # to not have prints during loading
                **params,
            )

        # All done
        log.success("Successfully loaded %d data entries.", len(load_cfg))

        # Finally, print the tree
        if print_tree:
            if print_tree == "condensed":
                print(self.tree_condensed)
            else:
                print(self.tree)

    def load(
        self,
        entry_name: str,
        *,
        loader: str,
        enabled: bool = True,
        glob_str: Union[str, List[str]],
        base_path: str = None,
        target_group: str = None,
        target_path: str = None,
        print_tree: Union[bool, str] = False,
        load_as_attr: bool = False,
        **load_params,
    ) -> None:
        """Performs a single load operation.

        Args:
            entry_name (str): Name of this entry; will also be the name of the
                created group or container, unless ``target_basename`` is given
            loader (str): The name of the loader to use
            enabled (bool, optional): Whether the load operation is enabled.
                If not, simply returns without loading any data or performing
                any further checks.
            glob_str (Union[str, List[str]]): A glob string or a list of glob
                strings by which to identify the files within ``data_dir`` that
                are to be loaded using the given loader function
            base_path (str, optional): The base directory to concatenate the
                glob string to; if None, will use the DataManager's data
                directory. With this option, it becomes possible to load data
                from a path outside the associated data directory.
            target_group (str, optional): If given, the files to be loaded will
                be stored in this group. This may only be given if the argument
                target_path is *not* given.
            target_path (str, optional): The path to write the data to. This
                can be a format string. It is evaluated for each file that has
                been matched. If it is not given, the content is loaded to a
                group with the name of this entry at the root level.
                Available keys are: ``basename``, ``match`` (if ``path_regex``
                is used, see ``**load_params``)
            print_tree (Union[bool, str], optional): If True, the full tree
                representation of the DataManager is printed after the data
                was loaded. If ``'condensed'``, the condensed tree will be
                printed.
            load_as_attr (bool, optional): If True, the loaded entry will be
                added not as a new DataContainer or DataGroup, but as an
                attribute to an (already existing) object at ``target_path``.
                The name of the attribute will be the ``entry_name``.
            **load_params: Further loading parameters, all optional. These are
                evaluated by :py:meth:`~dantro.data_mngr.DataManager._load`.

                ignore (list):
                    The exact file names in this list will be ignored during
                    loading. Paths are seen as elative to the data directory
                    of the data manager.
                required (bool):
                    If True, will raise an error if no files were found.
                    Default: False.
                path_regex (str):
                    This pattern can be used to match the path of the file
                    that is being loaded. The match result is available to the
                    format string under the ``match`` key.
                exists_action (str):
                    The behaviour upon existing data.
                    Can be: ``raise`` (default), ``skip``, ``skip_nowarn``,
                    ``overwrite``, ``overwrite_nowarn``.
                    With ``*_nowarn`` values, no warning is given if an entry
                    already existed. Note that this is ignored when
                    the ``load_as_attr`` argument is given.
                unpack_data (bool, optional):
                    If True, and ``load_as_attr`` is active, not the
                    DataContainer or DataGroup itself will be stored in the
                    attribute, but the content of its ``.data`` attribute.
                progress_indicator (bool):
                    Whether to print a progress indicator or not. Default: True
                parallel (bool):
                    If True, data is loaded in parallel. This feature is not
                    implemented yet!
                any further kwargs:
                    passed on to the loader function

        Returns:
            None

        Raises:
            ValueError: Upon invalid combination of ``target_group`` and
                ``target_path`` arguments
        """

        def glob_match_single(glob_str: Union[str, List[str]]) -> bool:
            """Returns True if the given glob str matches at most one file."""
            return bool(isinstance(glob_str, str) and glob_str.find("*") < 0)

        def check_target_path(target_path: str):
            """Check that the target path evaluates correctly."""
            log.debug("Checking target path '%s' ...", target_path)
            try:
                _target_path = target_path.format(
                    basename="basename", match="match"
                )

            except (IndexError, KeyError) as err:
                raise ValueError(
                    "Invalid argument `target_path`. Will not be able to "
                    f"properly evaluate '{target_path}' later due to "
                    f"a {type(err)}: {err}"
                ) from err
            else:
                log.debug("Target path will be:  %s", _target_path)

        # Initial checks
        if not enabled:
            log.progress("Skipping loading of data entry '%s' ...", entry_name)
            return
        log.progress("Loading data entry '%s' ...", entry_name)

        t0 = time.time()

        # Parse the arguments that result in the target path
        if load_as_attr:
            if not target_path:
                raise ValueError(
                    "With `load_as_attr`, the `target_path` "
                    "argument needs to be given."
                )

            # The target path should not be adjusted, as it points to the
            # object to store the loaded data as attribute in.
            log.debug(
                "Will load this entry as attribute to the target path "
                "'%s' ...",
                target_path,
            )

            # To communicate the attribute name, store it in the load_as_attr
            # variable; otherwise it would require passing two arguments to
            # _load
            load_as_attr = entry_name

        elif target_group:
            if target_path:
                raise ValueError(
                    "Received both arguments `target_group` and "
                    "`target_path`; make sure to only pass one "
                    "or none of them."
                )

            if glob_match_single(glob_str):
                target_path = target_group + "/" + entry_name
            else:
                target_path = target_group + "/{basename:}"

        elif not target_path:
            if glob_match_single(glob_str):
                target_path = entry_name
            else:
                target_path = entry_name + "/{basename:}"

        # else: target_path was given

        # ...and check that it is working.
        check_target_path(target_path)

        # Try loading the data and handle specific DataManagerErrors
        try:
            num_files = self._load(
                target_path=target_path,
                loader=loader,
                glob_str=glob_str,
                base_path=base_path,
                load_as_attr=load_as_attr,
                **load_params,
            )

        except RequiredDataMissingError:
            raise

        except MissingDataError as err:
            warnings.warn(
                "No files were found to import!\n" + str(err),
                MissingDataWarning,
            )
            return  # Does not raise, but does not save anything either

        except LoaderError:
            raise

        else:
            # Everything loaded as desired
            log.progress(
                "Loaded all data for entry '%s' in ~%.1fs.\n",
                entry_name,
                time.time() - t0,
            )

        # Done with this entry. Print tree, if desired.
        if print_tree == "condensed":
            print(self.tree_condensed)
        elif print_tree:
            print(self.tree)

    def _load(
        self,
        *,
        target_path: str,
        loader: str,
        glob_str: Union[str, List[str]],
        load_as_attr: Union[str, None],
        base_path: str = None,
        ignore: List[str] = None,
        required: bool = False,
        path_regex: str = None,
        exists_action: str = "raise",
        unpack_data: bool = False,
        progress_indicator: bool = True,
        parallel: bool = False,
        **loader_kwargs,
    ) -> int:
        """Helper function that loads a data entry to the specified path.

        Args:
            target_path (str): The path to load the result of the loader to.
                This can be a format string; it is evaluated for each file.
                Available keys are: basename, match (if ``path_regex`` is
                given)
            loader (str): The loader to use
            glob_str (Union[str, List[str]]): A glob string or a list of glob
                strings to match files in the data directory
            load_as_attr (Union[str, None]): If a string, the entry will be
                loaded into the object at ``target_path`` under a new attribute
                with this name.
            base_path (str, optional): The base directory to concatenate the
                glob string to; if None, will use the DataManager's data
                directory. With this option, it becomes possible to load data
                from a path outside the associated data directory.
            ignore (List[str], optional): The exact file names in this list
                will be ignored during loading. Paths are seen as relative to
                the data directory.
            required (bool, optional): If True, will raise an error if no files
                were found.
            path_regex (str, optional): The regex applied to the relative path
                of the files that were found. It is used to generate the name
                of the target container. If not given, the basename is used.
            exists_action (str, optional): The behaviour upon existing data.
                Can be: ``raise`` (default), ``skip``, ``skip_nowarn``,
                ``overwrite``, ``overwrite_nowarn``. With ``*_nowarn`` values,
                no warning is given if an entry already existed.
                Note that this is ignored if ``load_as_attr`` is given.
            unpack_data (bool, optional): If True, and ``load_as_attr`` is
                active, not the DataContainer or DataGroup itself will be
                stored in the attribute, but the content of its ``.data``
                attribute.
            progress_indicator (bool, optional): Whether to print a progress
                indicator or not
            parallel (bool, optional): If True, data is loaded in parallel -
                not implemented yet!
            **loader_kwargs: passed on to the loader function

        Raises:
            NotImplementedError: For ``parallel == True``
            ValueError: Bad ``path_regex``

        Returns:
            int: Number of files that data was loaded from
        """

        def resolve_loader(loader: str) -> Tuple[Callable, str, Callable]:
            """Resolves the loader function"""
            load_func_name = "_load_" + loader.lower()
            try:
                load_func = getattr(self, load_func_name)

            except AttributeError as err:
                raise LoaderError(
                    f"Loader '{loader}' was not available to {self.logstr}! "
                    "Make sure to use a mixin class that supplies "
                    f"the '{load_func_name}' loader method."
                ) from err
            else:
                log.debug("Resolved '%s' loader function.", loader)

            try:
                TargetCls = getattr(load_func, "TargetCls")

            except AttributeError as err:
                raise LoaderError(
                    f"Load function {load_func} misses required attribute "
                    "'TargetCls'. Check your mixin!"
                ) from err

            return load_func, load_func_name, TargetCls

        def create_files_list(
            *,
            glob_str: Union[str, List[str]],
            ignore: List[str],
            base_path: str = None,
            required: bool = False,
            sort: bool = False,
        ) -> list:
            """Create the list of file paths to load from.

            Internally, this uses a set, thus ensuring that the paths are
            unique. The set is converted to a list before returning.

            Args:
                glob_str (Union[str, List[str]]): The glob pattern or a list of
                    glob patterns
                ignore (List[str]): The list of files to ignore
                base_path (str, optional): The base path for the glob pattern;
                    use data directory, if not given.
                required (bool, optional): Will lead to an error being raised
                    if no files could be matched
                sort (bool, optional): If true, sorts the list before returning

            Returns:
                list: the file paths to load

            Raises:
                MissingDataError: If no files could be matched
                RequiredDataMissingError: If no files could be matched but were
                    required.
            """
            # Create a set to assure that all files are unique
            files = set()

            # Assure it is a list of strings
            if isinstance(glob_str, str):
                # Is a single glob string
                # Put it into a list to handle the same as the given arg
                glob_str = [glob_str]

            # Assuming glob_str to be lists of strings now
            log.debug(
                "Got %d glob string(s) to create set of matching file "
                "paths from.",
                len(glob_str),
            )

            # Handle base path, defaulting to the data directory
            if base_path is None:
                base_path = self.dirs["data"]
                log.debug("Using data directory as base path.")

            else:
                if not os.path.isabs(base_path):
                    raise ValueError(
                        "Given base_path argument needs be an "
                        f"absolute path, was not: {base_path}"
                    )

            # Go over the given glob strings and add to the files set
            for gs in glob_str:
                # Make the glob string absolute
                gs = os.path.join(base_path, gs)
                log.debug("Adding files that match glob string:\n  %s", gs)

                # Add to the set of files; this assures uniqueness of the paths
                files.update(list(glob.glob(gs, recursive=True)))

            # See if some files should be ignored
            if ignore:
                log.debug("Got list of files to ignore:\n  %s", ignore)

                # Make absolute and generate list of files to exclude
                ignore = [
                    os.path.join(self.dirs["data"], path) for path in ignore
                ]

                log.debug("Removing them one by one now ...")

                # Remove the elements one by one
                while ignore:
                    rmf = ignore.pop()
                    try:
                        files.remove(rmf)
                    except KeyError:
                        log.debug("%s was not found in set of files.", rmf)
                    else:
                        log.debug("%s removed from set of files.", rmf)

            # Now the file list is final
            log.note(
                "Found %d file%s to load.",
                len(files),
                "s" if len(files) != 1 else "",
            )
            log.debug("\n  %s", "\n  ".join(files))

            if not files:
                # No files found; exit here, one way or another
                if not required:
                    raise MissingDataError(
                        "No files found matching "
                        f"`glob_str` {glob_str} (and ignoring {ignore})."
                    )
                raise RequiredDataMissingError(
                    f"No files found matching `glob_str` {glob_str} "
                    f"(and ignoring {ignore}) were found, but were "
                    "marked as required!"
                )

            # Convert to list
            files = list(files)

            # Sort, if asked to do so
            if sort:
                files.sort()

            return files

        def prepare_target_path(
            target_path: str, *, filepath: str, path_sre=None
        ) -> List[str]:
            """Prepare the target path"""
            # The dict to be filled with formatting parameters
            fps = dict()

            # Extract the file basename (without extension)
            fps["basename"] = os.path.splitext(os.path.basename(filepath))[0]
            fps["basename"] = fps["basename"].lower()

            # Use the specified regex pattern to extract a match
            if path_sre:
                try:
                    _match = path_sre.findall(filepath)[0]

                except IndexError:
                    # nothing could be found
                    warnings.warn(
                        "Could not extract a name using the "
                        f"regex pattern '{path_sre}' on the file path:\n"
                        f"{filepath}\nUsing the path's basename instead.",
                        NoMatchWarning,
                    )
                    _match = fps["basename"]

                else:
                    log.debug(
                        "Matched '%s' in file path '%s'.", _match, filepath
                    )

                fps["match"] = _match

            # Parse the format string to generate the file path
            log.debug(
                "Parsing format string '%s' to generate target path ...",
                target_path,
            )
            log.debug("  kwargs: %s", fps)
            target_path = target_path.format(**fps)

            log.debug("Generated target path:  %s", target_path)
            return target_path.split(PATH_JOIN_CHAR)

        def skip_path(path: str, *, exists_action: str) -> bool:
            """Check whether a given path exists and — depending on the
            `exists_action` – decides whether to skip this path or now.

            Args:
                path (str): The path to check for existence.
                exists_action (str): The behaviour upon existing data. Can be:
                    raise, skip, skip_nowarn, overwrite, overwrite_nowarn.
                    The *_nowarn arguments suppress the warning

            Returns:
                bool: Whether to skip this path

            Raises:
                ExistingDataError: Raised when `exists_action == 'raise'`
                ValueError: Raised for invalid `exists_action` value
            """
            if path not in self:
                # Does not exist yet -> no need to skip
                return False
            # else: path exists already
            # NOTE that it is not known whether the path points to a group
            # or to a container

            _msg = "Path '{}' already exists.".format(
                PATH_JOIN_CHAR.join(path)
            )

            # Distinguish different actions
            if exists_action == "raise":
                raise ExistingDataError(
                    _msg
                    + " Adjust argument `exists_action` to allow skipping "
                    "or overwriting of existing entries."
                )

            if exists_action in ["skip", "skip_nowarn"]:
                if exists_action == "skip":
                    warnings.warn(
                        _msg + " Loading of this entry will be skipped.",
                        ExistingDataWarning,
                    )
                return True  # will lead to the data not being loaded

            elif exists_action in ["overwrite", "overwrite_nowarn"]:
                if exists_action == "overwrite":
                    warnings.warn(
                        _msg + " It will be overwritten!", ExistingDataWarning
                    )
                return False  # will lead to the data being loaded

            else:
                raise ValueError(
                    "Invalid value for `exists_action` argument "
                    f"'{exists_action}'! Can be: raise, skip, "
                    "skip_nowarn, overwrite, overwrite_nowarn."
                )

        def store(
            obj: Union[BaseDataGroup, BaseDataContainer],
            *,
            target_path: List[str],
            as_attr: Union[str, None],
            unpack_data: bool,
        ) -> None:
            """Store the given `obj` at the supplied `path`.

            Note that this will automatically overwrite, assuming that all
            checks have been made prior to the call to this function.

            Args:
                obj (Union[BaseDataGroup, BaseDataContainer]): Object to store
                target_path (List[str]): The path to store the object at
                as_attr (Union[str, None]): If a string, store the object in
                    the attributes of the container or group at target_path

            Raises:
                ExistingDataError: If non-group-like data already existed at
                    that path
                RequiredDataMissingError: If storing as attribute was selected
                    but there was no object at the given target_path
            """
            # First, handle the (easy) case where the object is to be stored
            # as the attribute at the target_path
            if as_attr:
                # Try to load the object at the target path
                try:
                    target = self[target_path]

                except KeyError as err:
                    raise RequiredDataMissingError(
                        f"In order to store the object {obj.logstr} at the "
                        f"target path '{target_path}', a group or container "
                        "already needs to exist at that location "
                        f"within {self.logstr}."
                    ) from err

                # Check whether an attribute with that name already exists
                if as_attr in target.attrs:
                    raise ExistingDataError(
                        f"An attribute with the name '{as_attr}' "
                        f"already exists in {target.logstr}!"
                    )

                # All checks passed. Can store it now, either directly or with
                # unpacking of its data ...
                if not unpack_data:
                    target.attrs[as_attr] = obj
                else:
                    target.attrs[as_attr] = obj.data

                log.debug(
                    "Stored %s as attribute '%s' of %s.",
                    obj.classname,
                    as_attr,
                    target.logstr,
                )

                # Done here. Return.
                return

            # Extract a target group path and a base name from path list
            group_path = target_path[:-1]
            basename = target_path[-1]

            # Resolve the target group object; create it if necessary
            # Need to check whether it is given at all. If not, write into the
            # data manager directly
            if not group_path:
                # Write directly into data manager root
                group = self

            else:
                # Need to retrieve or create the group
                # The difficulty is that the path can also point to a container
                # Need to assure here, that the group path points to a group
                if group_path not in self:
                    # Needs to be created
                    self._create_groups(group_path)

                elif not isinstance(self[group_path], BaseDataGroup):
                    # Already exists, but is no group. Cannot continue
                    group_path = PATH_JOIN_CHAR.join(group_path)
                    target_path = PATH_JOIN_CHAR.join(target_path)
                    raise ExistingDataError(
                        f"The object at '{group_path}' in {self.logstr} is "
                        f"not a group but a {type(self[group_path])}. Cannot "
                        f"store {obj.logstr} there because the target path "
                        f"'{target_path}' requires it to be a group."
                    )

                # Now the group path will point to a group
                group = self[group_path]

            # Store data, if possible
            if basename in group:
                # Already exists. Delete the old one, then store the new one
                del group[basename]

            # Can add now
            group.add(obj)

            # Done
            log.debug(
                "Successfully stored %s at '%s'.",
                _data.logstr,
                PATH_JOIN_CHAR.join(target_path),
            )

        # End of helper functions . . . . . . . . . . . . . . . . . . . . . . .
        # Get the loader function
        load_func, load_func_name, TargetCls = resolve_loader(loader)

        # Create the list of file paths to load
        files = create_files_list(
            glob_str=glob_str,
            ignore=ignore,
            required=required,
            base_path=base_path,
            sort=True,
        )

        # If a regex pattern was specified, compile it
        path_sre = re.compile(path_regex) if path_regex else None

        # Check if the `match` key is being used in the target_path
        if path_sre is not None and target_path.find("{match:") < 0:
            raise ValueError(
                "Received the `path_regex` argument to match the "
                "file path, but the `target_path` argument did "
                "not contain the corresponding `{{match:}}` "
                f"placeholder. `target_path` value: '{target_path}'."
            )

        if parallel:
            # TODO could be implemented by parallelising the below for loop
            raise NotImplementedError("Cannot load in parallel yet.")

        # Ready for loading files now . . . . . . . . . . . . . . . . . . . . .
        num_files = len(files)

        # Go over the files and load them
        for n, file in enumerate(files):
            self._progress_info_str = f"  Loading  {n+1}/{num_files}  ..."
            if progress_indicator:
                print(fill_line(self._progress_info_str), end="\r")

            # Prepare the target path (a list of strings)
            _target_path = prepare_target_path(
                target_path, filepath=file, path_sre=path_sre
            )

            # Distinguish regular loading and loading as attribute
            if not load_as_attr:
                # Check if it is to be skipped
                if skip_path(_target_path, exists_action=exists_action):
                    log.debug("Skipping file '%s' ...", file)
                    continue

                # Prepare the target class, which will be filled by the load
                # function; this assures that the name is already correct
                _TargetCls = lambda **kws: TargetCls(
                    name=_target_path[-1], **kws
                )

            else:
                # For loading as attribute, the exists_action is not valid;
                # that check is thus not needed. Also, the target class name
                # does not come from the target path but from that argument
                _TargetCls = lambda **kws: TargetCls(name=load_as_attr, **kws)

            # Get the data
            _data = load_func(file, TargetCls=_TargetCls, **loader_kwargs)
            log.debug(
                "Successfully loaded file '%s' into %s.", file, _data.logstr
            )

            # If this succeeded, store the data
            store(
                _data,
                target_path=_target_path,
                as_attr=load_as_attr,
                unpack_data=unpack_data,
            )

            # Done with this file. Go to next iteration

        # Clear the line to get rid of the load indicator, if there was one
        if progress_indicator:
            clear_line()

        # Done
        log.debug("Finished loading data from %d file(s).", len(files))
        return len(files)

    def _contains_group(
        self, path: Union[str, List[str]], *, base_group: BaseDataGroup = None
    ) -> bool:
        """Recursively checks if the given path is available _and_ a group.

        Args:
            path (Union[str, List[str]]): The path to check.
            base_group (BaseDataGroup): The group to start from. If not
                given, will use self.

        Returns:
            bool: Whether the path points to a group

        """

        def check(path: str, base_group: BaseDataGroup) -> bool:
            """Returns True if the object at path within base_group is
            a group. False otherwise.
            """
            return path in base_group and isinstance(
                base_group[path], BaseDataGroup
            )

        if not isinstance(path, list):
            path = path.split(PATH_JOIN_CHAR)

        if not base_group:
            base_group = self

        if len(path) > 1:
            # Need to continue recursively
            if check(path[0], base_group):
                return self._contains_group(
                    path[1:], base_group=base_group[path[0]]
                )
            return False

        # End of recursion
        return check(path[0], base_group)

    def _create_groups(
        self,
        path: Union[str, List[str]],
        *,
        base_group: BaseDataGroup = None,
        GroupCls: Union[type, str] = None,
        exist_ok: bool = True,
    ):
        """Recursively create groups for the given path. Unlike new_group, this
        also creates the groups at the intermediate paths.

        Args:
            path (Union[str, List[str]]): The path to create groups along
            base_group (BaseDataGroup, optional): The group to start from. If
                not given, uses self.
            GroupCls (Union[type, str], optional): The class to use for
                creating the groups or None if the _DATA_GROUP_DEFAULT_CLS is
                to be used. If a string is given, lookup happens from the
                _DATA_GROUPS_CLASSES variable.
            exist_ok (bool, optional): Whether it is ok that groups along the
                path already exist. These might also be of different type.
                Default: True

        Raises:
            ExistingDataError: If not `exist_ok`
            ExistingGroupError: If not `exist_ok` and a group already exists
        """
        # Parse arguments
        if isinstance(path, str):
            path = path.split(PATH_JOIN_CHAR)

        if base_group is None:
            base_group = self

        GroupCls = self._determine_group_class(GroupCls)

        # Catch the disallowed case as early as possible
        if path[0] in base_group:
            # Check if it is a group that exists there
            if isinstance(base_group[path[0]], BaseDataGroup):
                if not exist_ok:
                    raise ExistingGroupError(path[0])

            else:
                # There is data (that is not a group) existing at the path.
                # Cannot continue
                raise ExistingDataError(
                    f"Tried to create a group '{path[0]}' in "
                    f"{base_group.logstr}, but a container was already stored "
                    "at that path."
                )

        # Create the group, if it does not yet exist
        if path[0] not in base_group:
            log.debug(
                "Creating group '%s' in %s ...", path[0], base_group.logstr
            )
            base_group.new_group(path[0])

        # path[0] is now created
        # Check whether to continue recursion
        if len(path) > 1:
            # Continue recursion
            self._create_groups(
                path[1:], base_group=base_group[path[0]], GroupCls=GroupCls
            )

    def _determine_group_class(self, Cls: Union[type, str]) -> type:
        """Helper function to determine the type of a group from an argument.

        Args:
            Cls (Union[type, str]): If None, uses the _DATA_GROUP_DEFAULT_CLS.
                If a string, tries to extract it from the _DATA_GROUP_CLASSES
                class variable. Otherwise, assumes this is already a type.

        Returns:
            type: The group class to use

        Raises:
            KeyError: If the string class name was not registered
            ValueError: If no _DATA_GROUP_CLASSES variable was populated
        """
        if Cls is None:
            return self._DATA_GROUP_DEFAULT_CLS

        if isinstance(Cls, str):
            cls_name = Cls

            if not self._DATA_GROUP_CLASSES:
                raise ValueError(
                    "The class variable _DATA_GROUP_CLASSES is "
                    "empty; cannot look up class type by the "
                    "given name '{}'.".format(cls_name)
                )

            elif cls_name not in self._DATA_GROUP_CLASSES:
                raise KeyError(
                    "The given class name '{}' was not registered "
                    "with this {}! Available classes: {}"
                    "".format(
                        cls_name, self.classname, self._DATA_GROUP_CLASSES
                    )
                )

            # everything ok, retrieve the class type
            return self._DATA_GROUP_CLASSES[cls_name]

        # else: assume it is already a type and just return the given argument
        return Cls

    # .........................................................................
    # Dumping and restoring the DataManager

    def _parse_file_path(self, path: str, *, default_ext=None) -> str:
        """Parses a file path: if it is a relative path, makes it relative to
        the associated data directory. If a default extension is specified and
        the path does not contain one, that extension is added.

        This helper method is used as part of dumping and storing the data
        tree, i.e. in the :py:meth:`~dantro.data_mngr.DataManager.dump` and
        :py:meth:`~dantro.data_mngr.DataManager.restore` methods.
        """
        path = os.path.expanduser(path)
        if not os.path.isabs(path):
            path = os.path.join(self.dirs["data"], path)

        # Handle file extension, adding a default extension if none was given
        if default_ext:
            path, ext = os.path.splitext(path)
            path += ext if ext else default_ext

        return path

    def dump(self, *, path: str = None, **dump_kwargs) -> str:
        """Dumps the data tree to a new file at the given path, creating any
        necessary intermediate data directories.

        For restoring, use :py:meth:`~dantro.data_mngr.DataManager.restore`.

        Args:
            path (str, optional): The path to store this file at. If this is
                not given, use the default tree cache path that was set up
                during initialization.
                If it is given and a relative path, it is assumed relative to
                the data directory.
                If the path does not end with an extension, the ``.d3`` (read:
                "data tree") extension is automatically added.
            **dump_kwargs: Passed on to ``pkl.dump``

        Returns:
            str: The path that was used for dumping the tree file
        """
        if path:
            path = self._parse_file_path(path, default_ext=DATA_TREE_DUMP_EXT)
        else:
            path = self.tree_cache_path
        os.makedirs(os.path.dirname(path), exist_ok=True)

        log.progress("Dumping data tree of %s ...", self.logstr)
        with open(path, mode="x+b") as file:
            pkl.dump(self, file, **dump_kwargs)

        log.progress("Successfully stored data tree in cache file.")
        log.note("  Path:       %s", path)
        log.note("  File size:  %s", format_bytesize(os.path.getsize(path)))
        return path

    def restore(
        self, *, from_path: str = None, merge: bool = False, **load_kwargs
    ):
        """Restores the data tree from a dump.

        For dumping, use :py:meth:`~dantro.data_mngr.DataManager.dump`.

        Args:
            from_path (str, optional): The path to restore this DataManager
                from. If it is not given, uses the default tree cache path
                that was set up at initialization.
                If it is a relative path, it is assumed relative to the data
                directory. Take care to add the corresponding file extension.
            merge (bool, optional): If True, uses a recursive update to merge
                the current tree with the restored tree.
                If False, uses :py:meth:`~dantro.data_mngr.DataManager.clear`
                to clear the current tree and then re-populates it with the
                restored tree.
            **load_kwargs: Passed on to ``pkl.load``

        Raises:
            FileNotFoundError: If no file is found at the (expanded) path.
        """
        if from_path:
            path = self._parse_file_path(from_path)
        else:
            path = self.tree_cache_path

        if not os.path.isfile(path):
            raise FileNotFoundError(
                f"Could not restore {self.logstr} as there was no tree "
                f"cache file at '{path}'!"
            )

        log.progress("Restoring %s from data tree file ...", self.logstr)
        log.note("  Path:       %s", path)
        log.note("  File size:  %s", format_bytesize(os.path.getsize(path)))

        with open(path, mode="rb") as file:
            dm = pkl.load(file, **load_kwargs)

        if not merge:
            log.note("  Mode:       clear and load")
            self.clear()
        else:
            log.note("  Mode:       load and merge")

        self.recursive_update(dm)
        log.progress("Successfully restored the data tree.")

    # .........................................................................
    # Working with the data in the tree

    def new_group(self, path: str, *, Cls: Union[type, str] = None, **kwargs):
        """Creates a new group at the given path.

        This is a slightly advanced version of the new_group method of the
        BaseDataGroup. It not only adjusts the default type, but also allows
        more ways how to specify the type of the group to create.

        Args:
            path (str): Where to create the group. Note that the intermediates
                of this path need to already exist.
            Cls (Union[type, str], optional): If given, use this type to
                create the group. If a string is given, resolves the type from
                the _DATA_GROUP_CLASSES class variable. If None, uses the
                default data group type of the data manager.
            **kwargs: Passed on to Cls.__init__

        Returns:
            Cls: the created group
        """
        # Use helper function to parse the group class correctly
        Cls = self._determine_group_class(Cls)

        return super().new_group(path, Cls=Cls, **kwargs)



"""Custom dantro exception classes."""

import inspect
from difflib import get_close_matches as _get_close_matches
from typing import Tuple

# -----------------------------------------------------------------------------


class DantroError(Exception):
    """Base class for all dantro-related errors"""


class DantroWarning(UserWarning):
    """Base class for all dantro-related warnings"""


# General data tree warnings ..................................................


class UnexpectedTypeWarning(DantroWarning):
    """Given when there was an unexpected type passed to a data container."""


# General data tree errors ....................................................


class ItemAccessError(KeyError, IndexError):
    """Raised upon bad access via __getitem__ or similar magic methods.

    This derives from both native exceptions KeyError and IndexError as these
    errors may be equivalent in the context of the dantro data tree, which is
    averse to the underlying storage container.

    See :py:class:`~dantro.base.BaseDataGroup` for example usage.
    """

    def __init__(
        self,
        obj: "AbstractDataContainer",
        *,
        key: str,
        show_hints: bool = True,
        prefix: str = None,
        suffix: str = None,
    ):
        """Set up an ItemAccessError object, storing some metadata that is
        used to create a helpful error message.

        Args:
            obj (AbstractDataContainer): The object from which item access was
                attempted but failed
            key (str): The key with which ``__getitem__`` was called
            show_hints (bool, optional): Whether to show hints in the error
                message, e.g. available keys or "Did you mean ...?"
            prefix (str, optional): A prefix string for the error message
            suffix (str, optional): A suffix string for the error message

        Raises:
            TypeError: Upon ``obj`` without attributes ``logstr`` and ``path``;
                or ``key`` not being a string.
        """
        if not (hasattr(obj, "logstr") and hasattr(obj, "path")):
            raise TypeError(
                "ItemAccessError can only be used for objects that supply the "
                "`logstr` and `path` attributes, i.e. objects similar to "
                f"those comprising the dantro data tree! Got {type(obj)}."
            )

        if not isinstance(key, str):
            raise TypeError(
                "ItemAccessError `key` argument needs to be a "
                f"string, but was {type(key)} {repr(key)}!"
            )

        super().__init__(obj, key, show_hints, prefix, suffix)

    def __str__(self) -> str:
        """Parse an error message, using the additional information to give
        hints on where the error occurred and how it can be resolved.
        """
        obj, key, show_hints, prefix, suffix = self.args
        prefix = prefix + " " if prefix else ""
        suffix = suffix if suffix else ""

        # If possible, add a hint.
        hint = ""
        if show_hints and hasattr(obj, "keys") and hasattr(obj, "__len__"):
            if len(obj) == 0:
                hint = "This object is empty (== has zero length). "

            elif len(obj) <= 5:
                _keys = ", ".join(obj.keys())
                hint = f"Available keys:  {_keys} "

            else:
                matches = _get_close_matches(key, obj.keys(), n=5, cutoff=0.0)
                _matches = ", ".join(matches)
                hint = f"Did you mean: {_matches} ? "

        return (
            f"{prefix}"
            f"No item '{key}' available in {obj.logstr} @ {obj.path}! "
            f"{hint}{suffix}"
        )


# DataManager-related .........................................................


class DataManagerError(DantroError):
    """All DataManager exceptions derive from this one"""


class RequiredDataMissingError(DataManagerError):
    """Raised if required data was missing."""


class MissingDataError(DataManagerError):
    """Raised if data was missing, but is not required."""


class ExistingDataError(DataManagerError):
    """Raised if data already existed."""


class ExistingGroupError(DataManagerError):
    """Raised if a group already existed."""


class LoaderError(DataManagerError):
    """Raised if a data loader was not available"""


class MissingDataWarning(DantroWarning):
    """Used as warning instead of MissingDataError"""


class ExistingDataWarning(DantroWarning):
    """If there was data already existing ..."""


class NoMatchWarning(DantroWarning):
    """If there was no regex match"""


# Plotting-related ............................................................


class PlottingError(DantroError):
    """Custom exception class for all plotting errors"""


class PlotConfigError(ValueError, PlottingError):
    """Raised when there were errors in the plot configuration"""


class InvalidCreator(ValueError, PlottingError):
    """Raised when an invalid creator was specified"""


class PlotCreatorError(PlottingError):
    """Raised when an error occured in a plot creator"""


# PlotHelper . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .


class EnterAnimationMode(DantroError):
    """An exception that is used to convey to any
    :py:class:`~dantro.plot_creators.pcr_ext.ExternalPlotCreator` or derived
    creator that animation mode is to be entered instead of a regular
    single-file plot.

    It can and should be invoked via
    :py:meth:`~dantro.plot_creators._plot_helper.PlotHelper.enable_animation`.

    This exception can be raised from within a plot function to dynamically
    decide whether animation should happen or not. Its counterpart is
    :py:exc:`~dantro.exceptions.ExitAnimationMode`.
    """


class ExitAnimationMode(DantroError):
    """An exception that is used to convey to any
    :py:class:`~dantro.plot_creators.pcr_ext.ExternalPlotCreator` or derived
    creator that animation mode is to be exited and a regular single-file plot
    should be carried out.

    It can and should be invoked via
    :py:meth:`~dantro.plot_creators._plot_helper.PlotHelper.disable_animation`.

    This exception can be raised from within a plot function to dynamically
    decide whether animation should happen or not. Its counterpart is
    :py:exc:`~dantro.exceptions.ExitAnimationMode`.
    """


class PlotHelperError(PlotConfigError):
    """Raised upon failure to invoke a specific plot helper function, this
    custom exception type stores metadata on the helper invocation in order
    to generate a useful error message.
    """

    def __init__(
        self,
        upstream_error: Exception,
        *,
        name: str,
        params: dict,
        ax_coords: Tuple[int, int] = None,
    ):
        """Initializes a PlotHelperError"""
        self.err = upstream_error
        self.name = name
        self.ax_coords = ax_coords
        self.params = params

    def __str__(self):
        """Generates an error message for this particular helper"""
        _params = "\n".join(
            [f"      {k}: {repr(v)}" for k, v in self.params.items()]
        )
        _where = (
            "the figure" if not self.ax_coords else f"axis {self.ax_coords}"
        )
        return (
            f"'{self.name}' failed on {_where}!  "
            f"{self.err.__class__.__name__}: {self.err}\n"
            f"    Invocation parameters were:\n{_params}\n"
        )

    @property
    def docstring(self) -> str:
        """Returns the docstring of this helper function"""
        from .plot_creators import PlotHelper

        func = getattr(PlotHelper, "_hlpr_" + self.name)
        return f"{self.name}\n{'-'*len(self.name)}\n{inspect.getdoc(func)}\n"


class PlotHelperErrors(ValueError):
    """This custom exception type gathers multiple individual instances of
    :py:class:`~dantro.exceptions.PlotHelperError`.
    """

    def __init__(self, *errors, show_docstrings: bool = True):
        """Bundle multiple PlotHelperErrors together

        Args:
            *errors: The individual instances of
                :py:class:`~dantro.exceptions.PlotHelperError`
            show_docstrings (bool, optional): Whether to show docstrings in the
                error message.
        """
        self._errors = list()
        self._axes = set()
        self._num = len(errors)
        self._docstrings = dict()
        self.show_docstrings = show_docstrings

        for error in errors:
            self._errors.append(error)
            self._axes.add(error.ax_coords)
            self._docstrings[error.name] = error.docstring

    @property
    def errors(self):
        return self._errors

    def __str__(self) -> str:
        """Generates a combined error message for *all* registered errors"""
        s = (
            f"Encountered {self._num} error(s) "
            f"for {len(self._docstrings)} different plot helper(s) "
            f"on {len(self._axes)} different axes!\n\n"
        )

        s += "\n".join([f"-- {e}" for e in self._errors]) + "\n"

        if self.show_docstrings:
            s += "Relevant Docstrings\n"
            s += "===================\n\n"
            s += "\n\n".join(self._docstrings.values())

        return s



"""This sub-module implements the basic mixin classes that are required
in the dantro.base module"""

import contextlib
import logging
import sys
import warnings

from ..abc import PATH_JOIN_CHAR, AbstractDataProxy
from ..exceptions import UnexpectedTypeWarning

# Local constants
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------


class AttrsMixin:
    """This Mixin class supplies the `attrs` property getter and setter and
    the private `_attrs` attribute.

    Hereby, the setter function will initialize a BaseDataAttrs-derived object
    and store it as an attribute.
    This relays the checking of the correct attribute format to the actual
    BaseDataAttrs-derived class.

    For changing the class that is used for the attributes, an overwrite of the
    _ATTRS_CLS class variable suffices.
    """

    # The class attribute that the attributes will be stored to
    _attrs = None

    # Define the class to use for storing attributes
    _ATTRS_CLS = None

    @property
    def attrs(self):
        """The container attributes."""
        return self._attrs

    @attrs.setter
    def attrs(self, new_attrs):
        """Setter method for the container `attrs` attribute."""
        if self._ATTRS_CLS is None:
            raise ValueError(
                "Need to declare the class variable _ATTRS_CLS "
                "in order to use the AttrsMixin!"
            )

        # Perform the initialisation
        self._attrs = self._ATTRS_CLS(name="attrs", attrs=new_attrs)


class SizeOfMixin:
    """Provides the __sizeof__ magic method and attempts to take into account
    the size of the attributes.
    """

    def __sizeof__(self) -> int:
        """Returns the size of the data (in bytes) stored in this container's
        data and its attributes.

        Note that this value is approximate. It is computed by calling the
        ``sys.getsizeof`` function on the data, the attributes, the name and
        some caching attributes that each dantro data tree class contains.
        Importantly, this is not a recursive algorithm.

        Also, derived classes might implement further attributes that are not
        taken into account either. To be more precise in a subclass, create a
        specific __sizeof__ method and invoke this parent method additionally.

        For more information, see the documentation of ``sys.getsizeof``:

            https://docs.python.org/3/library/sys.html#sys.getsizeof
        """
        nbytes = sys.getsizeof(self._data)
        nbytes += sys.getsizeof(self._attrs)
        nbytes += sys.getsizeof(self._name)
        nbytes += sys.getsizeof(self._logstr)

        return nbytes


class LockDataMixin:
    """This Mixin class provides a flag for marking the data of a group or
    container as locked.
    """

    # Whether the data is regarded as locked. Note name-mangling here.
    __locked = False

    @property
    def locked(self) -> bool:
        """Whether this object is locked"""
        return self.__locked

    def lock(self):
        """Locks the data of this object"""
        self.__locked = True
        self._lock_hook()

    def unlock(self):
        """Unlocks the data of this object"""
        self.__locked = False
        self._unlock_hook()

    def raise_if_locked(self, *, prefix: str = None):
        """Raises an exception if this object is locked; does nothing otherwise"""
        if self.locked:
            raise RuntimeError(
                "{}Cannot modify {} because it was already "
                "marked locked."
                "".format(prefix + " " if prefix else "", self.logstr)
            )

    def _lock_hook(self):
        """Invoked upon locking."""
        pass

    def _unlock_hook(self):
        """Invoked upon unlocking."""
        pass


class BasicComparisonMixin:
    """Provides a (very basic) ``__eq__`` method to compare equality."""

    def __eq__(self, other) -> bool:
        """Evaluates equality by making the following comparisons: identity,
        strict type equality, and finally: equality of the ``_data`` and
        ``_attrs`` attributes, i.e. the *private* attribute. This ensures that
        comparison does not trigger any downstream effects like resolution of
        proxies.

        If types do not match exactly, ``NotImplemented`` is returned, thus
        referring the comparison to the other side of the ``==``.
        """
        if other is self:
            return True

        if type(other) is not type(self):
            return NotImplemented

        return self._data == other._data and self._attrs == other._attrs


class CollectionMixin:
    """This Mixin class implements the methods needed for being a Collection.

    It relays all calls forward to the data attribute.
    """

    def __contains__(self, key) -> bool:
        """Whether the given key is contained in the items."""
        return bool(key in self.data)

    def __len__(self) -> int:
        """The number of items."""
        return len(self.data)

    def __iter__(self):
        """Iterates over the items."""
        return iter(self.data)


class ItemAccessMixin:
    """This Mixin class implements the methods needed for getting, setting,
    and deleting items. It relays all calls forward to the data attribute, but
    if given a list (passed down from above), it extracts it
    """

    def __getitem__(self, key):
        """Returns an item."""
        return self.data[self._item_access_convert_list_key(key)]

    def __setitem__(self, key, val):
        """Sets an item."""
        self.data[self._item_access_convert_list_key(key)] = val

    def __delitem__(self, key):
        """Deletes an item"""
        del self.data[self._item_access_convert_list_key(key)]

    def _item_access_convert_list_key(self, key):
        """If given something that is not a list, just return that key"""
        if isinstance(key, list):
            if len(key) > 1:
                return tuple(key)
            return key[0]
        return key


class MappingAccessMixin(ItemAccessMixin, CollectionMixin):
    """Supplies all methods that are needed for Mapping access.

    All calls are relayed to the data attribute.
    """

    def keys(self):
        """Returns an iterator over the data's keys."""
        return self.data.keys()

    def values(self):
        """Returns an iterator over the data's values."""
        return self.data.values()

    def items(self):
        """Returns an iterator over data's (key, value) tuples"""
        return self.data.items()

    def get(self, key, default=None):
        """Return the value at ``key``, or ``default`` if ``key`` is not
        available.
        """
        return self.data.get(key, default)


class CheckDataMixin:
    """This mixin class extends a BaseDataContainer-derived class to check the
    provided data before storing it in the container.

    It implements a general _check_data method, overwriting the placeholder
    method in the BaseDataContainer, and can be controlled via class variables.

    .. note::

        This is not suitable for checking containers that are added to an
        object of a BaseDataGroup-derived class!

    Attributes:
        DATA_ALLOW_PROXY (bool): Whether to allow _all_ proxy types, i.e.
            classes derived from AbstractDataProxy
        DATA_EXPECTED_TYPES (tuple, None): Which types to allow. If None, all
            types are allowed.
        DATA_UNEXPECTED_ACTION (str): The action to take when an unexpected
            type was supplied. Can be: raise, warn, ignore
    """

    # Specify expected data types for this container class
    DATA_EXPECTED_TYPES = None  # as tuple or None (allow all)
    DATA_ALLOW_PROXY = False  # to check for AbstractDataProxy
    DATA_UNEXPECTED_ACTION = "warn"  # Can be: raise, warn, ignore

    def _check_data(self, data) -> None:
        """A general method to check the received data for its type

        Args:
            data: The data to check

        Raises:
            TypeError: If the type was unexpected and the action was 'raise'
            ValueError: Illegal value for DATA_UNEXPECTED_ACTION class variable

        Returns:
            None
        """
        if self.DATA_EXPECTED_TYPES is None:
            # All types allowed
            return

        # Compile tuple of allowed types
        expected_types = self.DATA_EXPECTED_TYPES

        if self.DATA_ALLOW_PROXY:
            expected_types += (AbstractDataProxy,)

        # Check for expected types
        if isinstance(data, expected_types):
            return

        # else: was not of the expected type
        # Create a base message
        msg = (
            f"Unexpected type {type(data)} for data passed to {self.logstr}! "
            f"Expected types are: {expected_types}."
        )

        # Handle according to the specified action
        if self.DATA_UNEXPECTED_ACTION == "raise":
            raise TypeError(msg)

        elif self.DATA_UNEXPECTED_ACTION == "warn":
            warnings.warn(
                f"{msg}\nInitialization will work, but be informed "
                "that there might be errors at runtime.",
                UnexpectedTypeWarning,
            )

        elif self.DATA_UNEXPECTED_ACTION == "ignore":
            log.debug(msg + " Ignoring ...")

        else:
            raise ValueError(
                f"Illegal value '{self.DATA_UNEXPECTED_ACTION}' for class "
                f"variable DATA_UNEXPECTED_ACTION of {self.classname}. "
                "Allowed values are: raise, warn, ignore"
            )


class DirectInsertionModeMixin:
    """A mixin class that provides a context manager, within which insertion
    into the mixed-in class (think: group or container) can happen more
    directly. This is useful in cases where more assumptions can be made about
    the to-be-inserted data, thus allowing to make fewer checks during
    insertion (think: duplicates, key order, etc.).

    .. note::

        This direct insertion mode is not (yet) part of the public interface,
        as it has to be evaluated how robust and error-prone it is.
    """

    __in_direct_insertion_mode = False

    @property
    def with_direct_insertion(self) -> bool:
        """Whether the class this mixin is mixed into is currently in direct
        insertion mode.
        """
        return self.__in_direct_insertion_mode

    @contextlib.contextmanager
    def _direct_insertion_mode(self, *, enabled: bool = True):
        """A context manager that brings the class this mixin is used in into
        direct insertion mode. While in that mode, the
        :py:meth:`~dantro.mixins.base.DirectInsertionModeMixin.with_direct_insertion`
        property will return true.

        This context manager additionally invokes two callback functions, which
        can be specialized to perform certain operations when entering or
        exiting direct insertion mode: *Before* entering,
        :py:meth:`~dantro.mixins.base.DirectInsertionModeMixin._enter_direct_insertion_mode`
        is called. *After* exiting,
        :py:meth:`~dantro.mixins.base.DirectInsertionModeMixin._exit_direct_insertion_mode`
        is called.

        Args:
            enabled (bool, optional): whether to actually use direct insertion
                mode. If False, will yield directly without setting the toggle.
                This is equivalent to a null-context.
        """
        if not enabled:
            self.__in_direct_insertion_mode = False
            yield
            return

        log.trace(
            "Entering direct insertion mode of %s @ %s ...",
            self.logstr,
            self.path,
        )

        # Perform the entering callback
        try:
            self._enter_direct_insertion_mode()

        except Exception as exc:
            raise RuntimeError(
                "Error in callback while entering direct insertion mode of "
                f"{self.logstr} @ {self.name}! {type(exc).__name__}: {exc}"
            ) from exc

        # Now inside direct insertion mode
        self.__in_direct_insertion_mode = True

        try:
            # Yield control to the with-context now
            yield

        finally:
            # Will end up here if there was an exception within the context.
            log.trace(
                "Exiting direct insertion mode of %s @ %s ...",
                self.logstr,
                self.path,
            )
            self.__in_direct_insertion_mode = False

            # NOTE Important to NOT have a return here or handle any other
            #      error, otherwise exceptions from the context are discarded.

        # Perform the exiting callback
        try:
            self._exit_direct_insertion_mode()

        except Exception as exc:
            raise RuntimeError(
                "Error in callback while exiting direct insertion mode of "
                f"{self.logstr} @ {self.name}! {type(exc).__name__}: {exc}"
            ) from exc

    def _enter_direct_insertion_mode(self):
        """Called after entering direct insertion mode; can be overwritten to
        attach additional behaviour.
        """

        pass

    def _exit_direct_insertion_mode(self):
        """Called before exiting direct insertion mode; can be overwritten to
        attach additional behaviour.
        """

        pass
